#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import textwrap
from qtpy import QtCore, QtWidgets
from libopensesame.exceptions import osexception
from libopensesame.oslogging import oslogger
from libqtopensesame.widgets.item_view_button import item_view_button
from libqtopensesame.widgets.tree_item_item import tree_item_item
from libqtopensesame.widgets.qtitem_splitter import qtitem_splitter
from libqtopensesame.widgets import header_widget
from libqtopensesame._input.pool_select import pool_select
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'qtitem', category=u'core')


def requires_init(fnc):

	"""
	desc:
		A decorator that makes sure that an item's controls are initialized
		before a function is called.
	"""

	def inner(self, *args, **kwargs):

		if self.container_widget is None:
			self.init_edit_widget()
		return fnc(self, *args, **kwargs)

	return inner


def wait_cursor(fnc):

	"""
	desc:
		A decorator that shows a waiting cursor during the function
	"""

	def inner(self, *args, **kwargs):

		if self.main_window.is_busy():
			return fnc(self, *args, **kwargs)
		self.main_window.set_busy(True)
		retval = fnc(self, *args, **kwargs)
		self.main_window.set_busy(False)
		return retval

	return inner


class qtitem(object):

	"""Base class for the GUI controls of other items"""

	initial_view = u'controls'
	label_align = u'right'
	help_url = None
	lazy_init = True

	def __init__(self):

		"""Constructor"""

		# The auto-widgets are stored in name -> (var, widget) dictionaries
		self.auto_line_edit = {}
		self.auto_combobox = {}
		self.auto_spinbox = {}
		self.auto_slider = {}
		self.auto_editor = {}
		self.auto_checkbox = {}
		# Lazy initialization means that the control widgets are initialized
		# only when they are shown for the first time. This dramatically speeds
		# up opening of files. However, with some items this is not currently
		# possible, because their widgets are also referred when they are not
		# shown. Currently this means the inline_script.
		if self.lazy_init:
			self.container_widget = None
		else:
			self.init_edit_widget()
		self._dirty = False
		self._shown = False
		self.lock = False
		self.maximized = False
		self.set_validator()
		oslogger.debug(u'created %s' % self.name)
		self.cached_script = self.to_string()

	@property
	def main_window(self):
		return self.experiment.main_window

	@property
	def overview_area(self):
		return self.experiment.main_window.ui.itemtree

	@property
	def theme(self):
		return self.experiment.main_window.theme

	@property
	def tabwidget(self):
		return self.experiment.main_window.tabwidget

	@property
	def console(self):
		return self.experiment.main_window.console

	@property
	def extension_manager(self):
		return self.experiment.main_window.extension_manager

	@property
	def default_description(self):
		return _(u'Default description')

	def open_tab(self, select_in_tree=True):

		"""
		desc:
			Opens the tab if it wasn't yet open, and switches to it.
		"""

		unsaved = self.main_window.unsaved_changes
		self.tabwidget.add(self.widget(), self.item_icon(), self.name)
		self.main_window.set_unsaved(unsaved)
		if select_in_tree:
			self.experiment.main_window.ui.itemtree.select_item(
				self.name,
				open_tab=False
			)

	def close_tab(self):

		"""
		desc:
			Closes the tab if it was open.
		"""

		self.tabwidget.remove(self.widget())

	def set_focus(self):

		"""
		desc:
			Gives focus to the most important widget.
		"""

		if hasattr(self, u'focus_widget') and self.focus_widget is not None:
			self.focus_widget.setFocus()

	def set_focus_widget(self, widget, override=False):

		"""
		desc:
			Sets the widget that receives focus when the tab is opened.

		arguments:
			widget:
				desc:	The widget to receive focus or `None` to reset.
				type:	[QWidget, NoneType]

		keywords:
			override:
				desc:	Indicates whether the focus widget should be changed if
						there already is a focus widget.
				type:	bool
		"""

		if override or not hasattr(self, u'focus_widget') or self.focus_widget is None:
			self.focus_widget = widget

	def update_item_icon(self):

		"""
		desc:
			Updates the item icon.
		"""

		self.tabwidget.set_icon(self.widget(), self.item_icon())
		self.experiment.items.set_icon(self.name, self.item_icon())
		self.header_item_icon.setPixmap(
			self.theme.qpixmap(self.item_icon(), 32))

	def item_icon(self):

		"""
		returns:
			desc:	The name of the item icon.
			type:	unicode
		"""

		return self.item_type

	def show_tab(self):

		"""
		desc:
			Is called when the tab becomes visible, and updated the contents.
		"""

		# The first time that the tab is shown, the view is determined. For
		# the controls view, nothing needs to be done, because this is the
		# default view after initialization.
		if not self._shown:
			self._shown = True
			self.main_window.ui.itemtree.inhibit_drag = True
			oslogger.debug(u'Invalid initial_view: %s' % self.initial_view)
			if self.initial_view == u'script':
				self.set_view_script()
			elif self.initial_view == u'split':
				self.set_view_split()
			elif self.initial_view != u'controls':
				oslogger.warning(u'Invalid initial_view: %s' % self.initial_view)
		self.extension_manager.fire(u'prepare_open_item', name=self.name)
		self.update_script()
		self.edit_widget()
		self.main_window.ui.itemtree.select_item(self.name, open_tab=False)
		self.extension_manager.fire(u'open_item', name=self.name)

	@requires_init
	def widget(self):

		"""
		returns:
			desc:	The widget that is added to the tabwidget.
			type:	QWidget
		"""

		return self.container_widget

	def init_edit_widget(self, stretch=True):

		"""
		desc:
			Builds the UI.

		keywords:
			stretch:
				desc:	Indicates whether a vertical stretch should be added to
						the bottom of the controls. This is necessary if the
						controls don't expand.
				type:	bool
		"""

		oslogger.debug(u'initializing controls for {}'.format(self.name))
		# Header widget
		self.header = header_widget.header_widget(self)
		self.header_hbox = QtWidgets.QHBoxLayout()
		self.header_item_icon = self.theme.qlabel(self.item_icon())
		self.header_hbox.addWidget(self.header_item_icon)
		self.header_hbox.addWidget(self.header)
		self.header_hbox.setContentsMargins(0, 0, 0, 0)
		self.header_hbox.setSpacing(12)
		# Maximize button
		self.button_toggle_maximize = QtWidgets.QPushButton(
			self.theme.qicon(u'view-fullscreen'), u'')
		self.button_toggle_maximize.setToolTip(_(u'Toggle pop-out'))
		self.button_toggle_maximize.setIconSize(QtCore.QSize(16, 16))
		self.button_toggle_maximize.clicked.connect(self.toggle_maximize)
		self.header_hbox.addWidget(self.button_toggle_maximize)
		# View button
		self.button_view = item_view_button(self)
		self.header_hbox.addWidget(self.button_view)
		# Help button
		self.button_help = QtWidgets.QPushButton(self.theme.qicon(u"help"), u"")
		self.button_help.setToolTip(
			_(u"Tell me more about the %s item") % self.item_type)
		self.button_help.setIconSize(QtCore.QSize(16, 16))
		self.button_help.clicked.connect(self.open_help_tab)
		self.header_hbox.addWidget(self.button_help)
		self.header_widget = QtWidgets.QWidget()
		self.header_widget.setLayout(self.header_hbox)
		# The edit_grid is the layout that contains the actual controls for the
		# items.
		self.edit_grid = QtWidgets.QFormLayout()
		if self.label_align == u'right':
			self.edit_grid.setLabelAlignment(QtCore.Qt.AlignRight)
		self.edit_grid.setFieldGrowthPolicy(
			QtWidgets.QFormLayout.FieldsStayAtSizeHint)
		self.edit_grid.setContentsMargins(0, 0, 0, 0)
		self.edit_grid.setVerticalSpacing(6)
		self.edit_grid.setHorizontalSpacing(12)
		self.edit_grid_widget = QtWidgets.QWidget()
		self.edit_grid_widget.setLayout(self.edit_grid)
		# The edit_vbox contains the edit_grid and the header widget
		self.edit_vbox = QtWidgets.QVBoxLayout()
		self.edit_vbox.addWidget(self.edit_grid_widget)
		self.edit_vbox.setContentsMargins(0, 0, 0, 0)
		self.edit_vbox.setSpacing(12)
		if stretch:
			self.edit_vbox.addStretch()
		self._edit_widget = QtWidgets.QWidget()
		self._edit_widget.setWindowIcon(self.theme.qicon(self.item_type))
		self._edit_widget.setLayout(self.edit_vbox)
		# The script widget is not initialized yet, to improve performance.
		# This will happen only when the script is shown.
		self._script_widget = None
		self._script_frame = QtWidgets.QWidget()
		# The container_widget is the top-level widget that is actually inserted
		# into the tab widget.
		self.splitter = qtitem_splitter(self)
		self.set_view_controls()
		self.splitter.splitterMoved.connect(self.splitter_moved)
		self.container_vbox = QtWidgets.QVBoxLayout()
		self.container_vbox.setContentsMargins(12, 12, 12, 12)
		self.container_vbox.setSpacing(18)
		self.container_vbox.addWidget(self.header_widget)
		self.container_vbox.addWidget(self.splitter)
		self.container_widget = QtWidgets.QWidget()
		self.container_widget.setLayout(self.container_vbox)
		self.container_widget.on_activate = self.show_tab
		self.container_widget.__item__ = self.name

	@wait_cursor
	def init_script_widget(self):

		from pyqode_extras.widgets import OpenSesameCodeEdit

		oslogger.debug(u'initializing script widget for {}'.format(self.name))
		self._script_widget = OpenSesameCodeEdit(self.main_window)
		self._script_widget.focusOutEvent = self._script_focus_out
		self.extension_manager.fire(
			u'register_editor',
			editor=self._script_widget
		)
		self._script_button = QtWidgets.QPushButton(
			self.theme.qicon(u'dialog-apply'),
			_(u'Apply')
		)
		self._script_button.is_apply_button = True
		self._script_button.clicked.connect(
			self.apply_script_changes_and_switch_view
		)
		self._script_button_layout = QtWidgets.QHBoxLayout()
		self._script_button_layout.setContentsMargins(0, 0, 0, 0)
		self._script_button_layout.addStretch()
		self._script_button_layout.addWidget(self._script_button)
		self._script_button_widget = QtWidgets.QWidget()
		self._script_button_widget.setLayout(self._script_button_layout)
		self._script_layout = QtWidgets.QVBoxLayout()
		self._script_layout.addWidget(self._script_button_widget)
		self._script_layout.addWidget(self._script_widget)
		self._script_layout.setAlignment(QtCore.Qt.AlignRight)
		self._script_layout.setContentsMargins(0, 12, 0, 0)
		self._script_layout.setSpacing(12)
		self._script_frame.setLayout(self._script_layout)
		self.update_script(use_cache=False)

	def _script_focus_out(self, event):

		"""
		desc:
			Is called when the script editor loses focus, so that changes
			can be applied.

		arguments:
			event: a QFocusEvent
		"""

		# When the apply button is pressed, the script is also applied. But this
		# also triggers a focusOutEvent. Hence, this check avoids applying the
		# script twice.
		w = self.main_window.focusWidget()
		if (
			not isinstance(w, QtWidgets.QPushButton) or
			not hasattr(w, 'is_apply_button')
		):
			self.apply_script_changes()

	def splitter_moved(self, pos, index):

		"""
		desc:
			Is called when the splitter handle is manually moved.

		arguments:
			pos:
				desc:	The splitter-handle position.
				type:	int
			index:
				desc:	The index of the splitter handle. Since there is only
						one handle, this is always 0.
				type:	int
		"""

		sizes = self.splitter.sizes()
		self.edit_size = sizes[0]
		self.script_size = sizes[1]
		if self.script_size == 0:
			self.button_view.set_view_icon(u'controls')
		elif self.edit_size == 0:
			self.button_view.set_view_icon(u'script')
		else:
			self.button_view.set_view_icon(u'split')

	def set_view_controls(self):

		"""
		desc:
			Puts the splitter in control view.
		"""

		self.splitter.setSizes([self.splitter.width(), 0])
		self.button_view.set_view_icon(u'controls')

	def set_view_script(self):

		"""
		desc:
			Puts the splitter in script view.
		"""

		if self._script_widget is None:
			self.init_script_widget()
		self.splitter.setSizes([0, self.splitter.width()])
		self.button_view.set_view_icon(u'script')

	def set_view_split(self):

		"""
		desc:
			Puts the splitter in split view.
		"""

		if self._script_widget is None:
			self.init_script_widget()
		width = self.splitter.width() / 2
		self.splitter.setSizes([width, width])
		self.button_view.set_view_icon(u'split')

	def update(self):

		"""
		desc:
			Updates both the script and the controls.
		"""

		# Items are updated when their tab is shown, so we don't need to update
		# them if they aren't shown.
		if self.tabwidget.current_item() != self.name:
			return False
		self.update_script()
		self.edit_widget()
		return True

	def update_script(self, use_cache=True):

		"""
		desc:
			Regenerates the script and updates the script widget.
		"""

		script = self.to_string()
		if use_cache and self.cached_script == script:
			return  # nothing changed
		self.cached_script = script
		# Normally, the script starts with a 'define' line and is indented by
		# a tab. We want to undo this, and present only unindented content.
		script = script[script.find(u'\t'):]
		script = textwrap.dedent(script)
		self.main_window.set_unsaved()
		if self._script_widget is not None:
			self._script_widget.setPlainText(
				script,
				u'text/generic',
				u'utf-8'
			)
		self.extension_manager.fire(u'change_item', name=self.name)

	def edit_widget(self, *deprecated, **_deprecated):

		"""
		desc:
			This function updates the controls based on the item state.
		"""

		self.auto_edit_widget()
		self.header.refresh()

	def apply_edit_changes(self, *deprecated, **_deprecated):

		"""
		desc:
			Applies changes to the graphical controls.
		"""

		self.auto_apply_edit_changes()
		self.update_script()
		self.set_clean()
		return True

	def apply_script_changes(self, *deprecated, **_deprecated):

		"""
		desc:
			Applies changes to the script.
		"""

		if not self.validate_script():
			self.set_view_script()
			return
		new_script = self._script_widget.toPlainText()
		old_script = self.to_string()
		self.from_string(new_script)
		if old_script != new_script:
			self.main_window.set_unsaved()
			self.cached_script = new_script
		self.edit_widget()

	def apply_script_changes_and_switch_view(self, *deprecated, **_deprecated):

		"""
		desc:
			Applies changes to the script if possible. If so, switches to the
			controls view.
		"""

		if not self.validate_script():
			self.set_view_script()
			return
		self.apply_script_changes()
		self.set_view_controls()

	def validate_script(self):

		"""
		desc:
			Checks whether the script is syntactically valid. If not, the
			offending line is highlighted if QProgEdit suppors setInvalid().

		returns:
			type:	bool
		"""

		script = self._script_widget.toPlainText()
		# First create a dummy item to see if the string can be parsed.
		try:
			self.validator(self.name, self.experiment, script)
			return True
		except Exception as e:
			# If an error occurs, we first parse the first line, then the first
			# and second, and so on, until we find the error.
			l = script.split(u'\n')
			for line_nr, line in enumerate(l):
				test_script = u'\n'.join(l[:line_nr])
				try:
					self.validator(self.name, self.experiment, test_script)
				except Exception as e_:
					if not isinstance(e_, osexception):
						e_ = osexception(e_)
					from pyqode.core.api.utils import TextHelper
					TextHelper(self._script_widget).select_lines(
						line_nr - 1,
						line_nr - 1
					)
					break
			self.console.write(e)
			self.extension_manager.fire(
				u'notify',
				message=_(u'Failed to parse script. See debug window for details'),
				category=u'warning',
				always_show=True
			)
			return False

	def set_validator(self):

		"""
		desc:
			Sets the validator class, that is, the class that is used to parse
			scripts to see if they are syntactically correct. Currently, we use
			the class that defines from_string().
		"""

		import inspect
		meth = self.from_string
		for cls in inspect.getmro(self.__class__):
			if meth.__name__ in cls.__dict__:
				break
		oslogger.debug(u'validator: %s' % cls)
		self.validator = cls

	def rename(self, from_name, to_name):

		"""
		desc:
			Handles renaming of an item (not necesarrily the current item).

		arguments:
			from_name:
				desc:	The old item name.
				type:	unicode
			to_name:
				desc:	The new item name
				type:	unicode
		"""

		if self.name != from_name:
			return
		self.name = to_name
		if self.container_widget is None:
			return
		self.container_widget.__item__ = self.name
		self.header.set_name(to_name)
		index = self.tabwidget.indexOf(self.widget())
		if index is not None:
			self.tabwidget.setTabText(index, to_name)

	def open_help_tab(self):

		"""
		desc:
			Opens a help tab.
		"""

		if self.help_url is None:
			self.tabwidget.open_help(self.item_type)
		else:
			self.tabwidget.open_osdoc(self.help_url)

	def toggle_maximize(self):

		"""
		desc:
			Toggles edit-widget maximization.
		"""

		if not self.maximized:
			# Always ignore close events. This is necessary, because otherwise
			# the pop-out widget can be closed without re-enabling the main
			# window.
			self.main_window.block_close_event = True
			self.container_widget.closeEvent = lambda e: e.ignore()
			self.container_widget.setParent(None)
			self.container_widget.setWindowFlags(QtCore.Qt.WindowStaysOnTopHint|\
				QtCore.Qt.WindowMaximizeButtonHint|\
				QtCore.Qt.CustomizeWindowHint)
			self.container_widget.showMaximized()
			self.container_widget.show()
			self.button_toggle_maximize.setIcon(
				self.theme.qicon(u'view-restore'))
		else:
			self.main_window.block_close_event = False
			self.container_widget.setParent(self.main_window)
			self.open_tab()
			self.button_toggle_maximize.setIcon(
				self.theme.qicon(u'view-fullscreen'))
		self.maximized = not self.maximized
		self.button_help.setDisabled(self.maximized)
		self.main_window.setDisabled(self.maximized)

	def delete(self, item_name, item_parent=None, index=None):

		"""
		desc:
			Is called when an item is deleted (not necessarily the current one).

		arguments:
			item_name:
				desc:	The name of the item to be deleted.
				type:	str

		keywords:
			item_parent:
				desc:	The name of the parent item.
				type:	str
			index:
				desc:	The index of the item in the parent.
				type:	int
		"""

		pass

	def build_item_tree(
		self,
		toplevel=None,
		items=[],
		max_depth=-1,
		extra_info=None
	):

		"""
		desc:
			Constructs an item tree.

		keywords:
			toplevel:
				desc:	The toplevel widget.
				type:	tree_base_item
			items:
				desc:	A list of item names that have already been added, to
						prevent recursion.
				tyep:	list

		returns:
			type:	tree_item_item
		"""

		widget = tree_item_item(self, extra_info=extra_info)
		items.append(self.name)
		if toplevel is not None:
			toplevel.addChild(widget)
		return widget

	def parents(self):

		"""
		returns:
			desc:	A list of all parents (names) of the current item.
			type:	list
		"""

		l = [self.name]
		for item in self.experiment.items:
			if self.experiment.items[item].is_child_item(self.name):
				l.append(item)
		return l

	def get_ready(self):

		"""
		desc:
			This function should be overridden to do any last-minute stuff that
			and item should do before an experiment is actually run, such as
			applying pending script changes.

		returns:
			desc:	True if some action has been taken, False if nothing was
					done.
			type:	bool
		"""

		if self._dirty:
			self.apply_edit_changes()
			return True
		return False

	def auto_edit_widget(self):

		"""
		desc:
			Updates the GUI controls based on the auto-widgets.
		"""

		for var, edit in self.auto_line_edit.items():
			if isinstance(var, int):
				continue
			if var in self.var:
				val = safe_decode(self.var.get(var, _eval=False))
			else:
				val = u''
			if val != edit.text():
				edit.setText(val)

		for var, combobox in self.auto_combobox.items():
			if isinstance(var, int):
				continue
			val = self.var.get(var, _eval=False, default=u'')
			i = combobox.findText(safe_decode(val))
			# Set the combobox to the select item
			if i >= 0:
				if i != combobox.currentIndex() or not combobox.isEnabled():
					combobox.setDisabled(False)
					combobox.setCurrentIndex(i)
			# If no value was specified, set the combobox to a blank item
			elif val == u'':
				if combobox.currentIndex() >= 0 or not combobox.isEnabled():
					combobox.setDisabled(False)
					combobox.setCurrentIndex(-1)
			# If an unknown value has been specified, notify the user
			else:
				if combobox.isEnabled():
					combobox.setDisabled(True)
					self.extension_manager.fire(u'notify',
						message=_(u'"%s" is set to a variable or '
						u'unknown value and can only be edited through '
						u'the script.') % var, category=u'info')

		for var, spinbox in list(self.auto_spinbox.items()) \
			+ list(self.auto_slider.items()):
			if isinstance(var, int):
				continue
			if var in self.var:
				val = self.var.get(var, _eval=False)
				if type(val) in (float, int):
					if spinbox.value() == val and spinbox.isEnabled():
						continue
					spinbox.setDisabled(False)
					try:
						spinbox.setValue(val)
					except Exception as e:
						self.experiment.notify(
							_(u"Failed to set control '%s': %s") % (var, e))
				else:
					if not spinbox.isEnabled():
						continue
					spinbox.setDisabled(True)
					self.extension_manager.fire(u'notify',
						message=_(u'"%s" is defined using '
						'variables and can only be edited through the '
						'script.') % var, category=u'info')

		for var, checkbox in self.auto_checkbox.items():
			if isinstance(var, int):
				continue
			if var in self.var:
				val = self.var.get(var, _eval=False)
				if val in [u'yes', u'no']:
					checkbox.setDisabled(False)
					checked = val == u'yes'
					if checked != checkbox.isChecked():
						checkbox.setChecked(checked)
				else:
					checkbox.setDisabled(True)
					self.extension_manager.fire(u'notify',
						message=_(u'"%s" is defined using '
						u'variables or has an invalid value, and can only be '
						u'edited through the script.') % var,
						category=u'info')

		for var, editor in self.auto_editor.items():
			if isinstance(var, int):
				continue
			val = safe_decode(self.var.get(var, _eval=False, default=u''))
			if val != editor.toPlainText():
				editor.setPlainText(val)

	def auto_apply_edit_changes(self):

		"""
		desc:
			Applies the auto-widget controls.
		"""

		for var, edit in self.auto_line_edit.items():
			if isinstance(var, int):
				continue
			if edit.isEnabled() and isinstance(var, basestring):
				val = edit.text().strip()
				if val:
					self.var.set(var, val)
					continue
				# If no text was entered, we use a default if available ...
				if hasattr(edit, u'default'):
					self.var.set(var, edit.default)
					continue
				# ... or unset the variable if no default is available.
				self.var.unset(var)

		for var, combobox in self.auto_combobox.items():
			if isinstance(var, int):
				continue
			if combobox.isEnabled() and isinstance(var, basestring):
				self.var.set(var, combobox.currentText())

		for var, spinbox in self.auto_spinbox.items():
			if isinstance(var, int):
				continue
			if spinbox.isEnabled() and isinstance(var, basestring):
				self.var.set(var, spinbox.value())

		for var, slider in self.auto_slider.items():
			if isinstance(var, int):
				continue
			if slider.isEnabled() and isinstance(var, basestring):
				self.var.set(var, slider.value())

		for var, checkbox in self.auto_checkbox.items():
			if isinstance(var, int):
				continue
			if checkbox.isEnabled() and isinstance(var, basestring):
				val = u'yes' if checkbox.isChecked() else u'no'
				self.var.set(var, val)

		for var, editor in self.auto_editor.items():
			if isinstance(var, int):
				continue
			if isinstance(var, basestring):
				self.var.set(var, editor.toPlainText())
		return True

	def auto_add_widget(self, widget, var=None, apply_func=None):

		"""
		desc:
			Adds a widget to the list of auto-widgets.

		arguments:
			widget:
			 	type:	QWidget

		keywords:
			var:	The name of the experimental variable to be linked to the
					widget, or None to use an automatically chosen name.
			type:	[str, NoneType]
		"""

		# Use the object id as a fallback name
		if var is None:
			var = id(widget)
		if apply_func is None:
			apply_func = self.apply_edit_changes
		self.set_focus_widget(widget)
		if (
			isinstance(widget, QtWidgets.QSpinBox) or
			isinstance(widget, QtWidgets.QDoubleSpinBox)
		):
			widget.editingFinished.connect(apply_func)
			widget.valueChanged.connect(self.set_dirty)
			self.auto_spinbox[var] = widget
		elif isinstance(widget, QtWidgets.QComboBox):
			widget.activated.connect(apply_func)
			self.auto_combobox[var] = widget
		elif (
			isinstance(widget, QtWidgets.QSlider) or
			isinstance(widget, QtWidgets.QDial)
		):
			widget.valueChanged.connect(apply_func)
			self.auto_slider[var] = widget
		elif (
			isinstance(widget, QtWidgets.QLineEdit) or
			isinstance(widget, pool_select)
		):
			widget.editingFinished.connect(apply_func)
			widget.textEdited.connect(self.set_dirty)
			self.auto_line_edit[var] = widget
		elif isinstance(widget, QtWidgets.QCheckBox):
			widget.clicked.connect(apply_func)
			self.auto_checkbox[var] = widget
		else:
			raise TypeError(u"Cannot auto-add widget of type %s" % widget)

	def set_dirty(self):

		"""
		desc:
			Indicates that the item's controls have changed since being last
			applied.
		"""

		self._dirty = True

	def set_clean(self):

		"""
		desc:
			Indicates that the item's controls have all been applied.
		"""

		self._dirty = False

	def children(self):

		"""
		returns:
			desc:	A list of children, including grand children, and so on.
			type:	list
		"""

		return []
		
	def direct_children(self):
		
		"""
		returns:
			desc:	A list of direct children, which does not include grand
					children, and so on.
			type:	list
		"""		
		
		return []

	def is_child_item(self, item_name):

		"""
		desc:
			Checks if an item is somewhere downstream from the current item
			in the experimental hierarchy.

		arguments:
			item_name:
				desc:	The name of the child item.
				type:	unicode

		returns:
			desc:	True if the current item is offspring of the item, False
					otherwise.
			type:	bool
		"""

		return False

	def insert_child_item(self, item_name, index=0):

		"""
		desc:
			Inserts a child item, if applicable to the item type.

		arguments:
			item_name:
				desc:	The name of the child item.
				type:	unicode

		keywords:
			index:
				desc:	The index of the child item.
				type:	int
		"""

		pass

	def remove_child_item(self, item_name, index=0):

		"""
		desc:
			Removes a child item, if applicable to the item type.

		arguments:
			item_name:
				desc:	The name of the child item.
				type:	unicode

		keywords:
			index:
				desc:	The index of the child item, if applicable. A negative
						value indicates all instances.
				type:	int
		"""

		pass


QtItem = qtitem



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from openexp._coordinates.coordinates import Coordinates


class Xpyriment(Coordinates):

	"""
	desc:
		For function specifications and docstrings, see
		`openexp._coordinates.coordinates`.
	"""

	def __init__(self):

		Coordinates.__init__(self)
		self._xwcenter = self.experiment.expyriment.screen.window_size[0]/2
		self._ywcenter = self.experiment.expyriment.screen.window_size[1]/2

	def to_xy(self, x, y=None):

		if isinstance(x, tuple):
			x, y = x
		x, y = self.none_to_center(x, y)
		if self._canvas_dev:
			# For expyriment, 0,0 is the display center and positive y
			# coordinates are up.
			if self.uniform_coordinates:
				return x, -y
			return x - self._xcenter, self._ycenter - y
		if self._mouse_dev:
			# The mouse is centered on the top-left, but we need to take into
			# account that the display is padded in fullscreen mode.
			if self.uniform_coordinates:
				return x + self._xwcenter, y + self._ywcenter
			return x + self._xwcenter - self._xcenter, \
				y + self._ywcenter - self._ycenter

	def from_xy(self, x, y=None):

		if isinstance(x, tuple):
			x, y = x
		if not self._mouse_dev:
			raise osexception(u'Only mouse supported')
		if self.uniform_coordinates:
			return x - self._xwcenter, y - self._ywcenter
		return x - self._xwcenter + self._xcenter, \
			y - self._ywcenter + self._ycenter


# Non PEP-8 alias for backwards compatibility
xpyriment = Xpyriment



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from libopensesame import item
from libqtopensesame.items.qtautoplugin import qtautoplugin
import random


class advanced_delay(item.item):

	description = u'Waits for a specified duration'

	def reset(self):

		"""
		desc:
			Initialize plug-in.
		"""

		self.var.duration = 1000
		self.var.jitter = 0
		self.var.jitter_mode = u'Uniform'

	def prepare(self):

		"""
		desc:
			The preparation phase of the plug-in.
		"""

		item.item.prepare(self)
		# Sanity check on the duration value, which should be a positive numeric
		# value.
		if type(self.var.duration) not in (int, float) or self.var.duration < 0:
			raise osexception(
				u'Duration should be a positive numeric value in advanced_delay %s' \
				% self.name)
		if self.var.jitter_mode == u'Uniform':
			self._duration = random.uniform(self.var.duration-self.var.jitter/2,
				self.var.duration+self.var.jitter/2)
		elif self.var.jitter_mode == u'Std. Dev.':
			self._duration = random.gauss(self.var.duration, self.var.jitter)
		else:
			raise osexception(
				u'Unknown jitter mode in advanced_delay %s' % self.name)
		# Don't allow negative durations.
		if self._duration < 0:
			self._duration = 0
		self._duration = int(self._duration)
		self.experiment.var.set(u'delay_%s' % self.name, self._duration)

	def run(self):

		"""
		desc:
			The run phase of the plug-in.
		"""

		self.set_item_onset(self.time())
		self.sleep(self._duration)

	def var_info(self):

		"""
		Gives a list of dictionaries with variable descriptions.

		Returns:
		A list of (name, description) tuples.
		"""

		return item.item.var_info(self) + [(u'delay_%s' % self.name, \
			u'[Determined at runtime]')]

class qtadvanced_delay(advanced_delay, qtautoplugin):

	"""Automatic plug-in GUI."""

	def __init__(self, name, experiment, script=None):

		advanced_delay.__init__(self, name, experiment, script)
		qtautoplugin.__init__(self, __file__)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import functools
from qtpy.QtWidgets import (
	QCheckBox,
	QSpinBox,
	QDoubleSpinBox,
	QKeySequenceEdit,
	QLineEdit,
	QComboBox,
	QFontComboBox
)
from qtpy.QtGui import QKeySequence, QFont
from libopensesame.oslogging import oslogger
from libqtopensesame.widgets.base_widget import BaseWidget
from libqtopensesame.misc.config import cfg


class BasePreferencesWidget(BaseWidget):
	
	def __init__(self, main_window, ui):

		"""
		desc:
			Constructor.

		arguments:
			main_window:	A qtopensesame object.
			ui:			A ui file.
		"""

		super(BasePreferencesWidget, self).__init__(main_window, ui=ui)
		self._its_me = False
		self._before_init_widgets()
		self._init_widgets()
		self._after_init_widgets()
		self.extension_manager.register_extension(self)
		
	def _before_init_widgets(self):
		
		"""
		desc:
			Can be implemented to perform some actions before the rest of
			the widgets are initialized.
		"""
		
		pass

	def _after_init_widgets(self):
		
		"""
		desc:
			Can be implemented to perform some actions after the rest of
			the widgets have been initialized.
		"""		
		
		pass

	def _init_widgets(self):
		
		"""
		desc:
			Introspects the widgets and automatically binds them to settings.
		"""
		
		for name in dir(self.ui):
			if not name.startswith('cfg_'):
				continue
			setting = name[4:]
			if setting not in cfg:
				oslogger.warning('unknown setting {}'.format(setting))
			widget = getattr(self.ui, name)
			
			def change_setting(setting, value):
				if self._its_me:
					return
				if isinstance(value, QKeySequence):
					value = value.toString()
				elif isinstance(value, QFont):
					value = value.family()
				elif isinstance(value, QLineEdit):
					value = value.text()
				cfg[setting] = value
				self._its_me = True
				self.extension_manager.fire(
					'setting_changed',
					setting=setting,
					value=value
				)
				self._its_me = False

			change_setting = functools.partial(change_setting, setting)
			if isinstance(widget, QCheckBox):
				widget.setChecked(cfg[setting])
				widget.toggled.connect(change_setting)
			elif isinstance(widget, QLineEdit):
				widget.setText(cfg[setting])
				widget.editingFinished.connect(
					functools.partial(change_setting, widget)
				)
			elif isinstance(widget, (QSpinBox, QDoubleSpinBox)):
				widget.setValue(cfg[setting])
				widget.valueChanged.connect(change_setting)
			elif isinstance(widget, QKeySequenceEdit):
				widget.setKeySequence(QKeySequence(cfg[setting]))
				widget.keySequenceChanged.connect(change_setting)
			elif isinstance(widget, QFontComboBox):
				widget.setCurrentFont(QFont(cfg[setting]))
				widget.currentFontChanged.connect(change_setting)
			elif isinstance(widget, QComboBox):
				widget.setCurrentText(cfg[setting])
				widget.currentTextChanged.connect(change_setting)
			else:
				oslogger.warning('invalid QWidget with name {}'.format(name))
				
	def event_setting_changed(self, setting, value):
		
		"""
		desc:
			Update the controls if settings have been changed by someone
			else.
			
		arguments:
			setting:	The name of the changed setting
			value:	The new value
		"""
		
		if self._its_me:
			return
		name = 'cfg_{}'.format(setting)
		if not hasattr(self.ui, name):
			return
		self._its_me = True
		widget = getattr(self.ui, name)
		if isinstance(widget, QCheckBox):
			widget.setChecked(bool(value))
		elif isinstance(widget, QLineEdit):
			widget.setText(value)
		elif isinstance(widget, (QSpinBox, QDoubleSpinBox)):
			try:
				widget.setValue(value)
			except TypeError:
				oslogger.warning(
					'invalid value {} for {}',format(value, setting)
				)
		elif isinstance(widget, QKeySequenceEdit):
			try:
				widget.setKeySequence(QKeySequence(value))
			except TypeError:
				oslogger.warning(
					'invalid value {} for {}',format(value, setting)
				)
		elif isinstance(widget, QFontComboBox):
			try:
				widget.setCurrentFont(QFont(value))
			except TypeError:
				oslogger.warning(
					'invalid value {} for {}',format(value, setting)
				)
		elif isinstance(widget, QComboBox):
			try:
				widget.setCurrentText(value)
			except TypeError:
				oslogger.warning(
					'invalid value {} for {}',format(value, setting)
				)
		else:
			oslogger.warn('invalid QWidget with name {}'.format(name))
		self._its_me = False



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from libopensesame import item, generic_response
from libopensesame.oslogging import oslogger
from libqtopensesame import qtplugin
try:
	from ctypes import windll
except:
	oslogger.warning(
		"failed to load ctypes.windll mode (only dummy mode will be available)"
	)


class port_reader(item.item, generic_response.generic_response):

	"""A plug-in to collect responses from a port (Window only)"""

	description = \
		u"Collects input from a port for the purpose of response collection"

	def reset(self):

		"""item.item."""

		self.var.timeout = "infinite"
		self.var.port = 889
		self.var.resting_value = 127
		self.var.auto_response = 255 # 'a'
		self.var.duration = "portinput"
		self.var.dummy = "no"
		self.process_feedback = True

	def get_portinput(self):

		"""Read port input"""

		while True:
			val = self._port.Inp32(self.port)
			time = self.time()
			if val != self.resting_value:
				break
			if self._timeout is not None and time - self.sri > self._timeout:
				val = "timeout"
				break
		return val, time

	def prepare_duration_portinput(self):

		"""Prepare a portinput duration"""

		if self.experiment.auto_response:
			self._duration_func = self.sleep_for_duration
			self._duration = 500
		else:
			try:
				self._port = windll.inpout32
			except:
				raise osexception( \
					"Failed to load inpout32.dll in port_reader '%s'" \
					% self.name)
			self._duration_func = self.get_portinput

	def process_response_portinput(self, retval):

		"""Process a portinput response"""

		self.experiment._start_response_interval = self.sri
		self.experiment.response, self.experiment.end_response_interval = retval

	def prepare(self):

		"""
		Prepare the item

		Returns:
		True on success, False on failure
		"""

		if self.dummy == "yes":
			self.duration = "keypress"

		item.item.prepare(self)
		generic_response.generic_response.prepare(self)
		return True

	def run(self):

		"""
		Runs the item

		Returns:
		True on success, False on failure
		"""

		# Record the onset of the current item
		self.set_item_onset()
		self.set_sri()
		self.process_response()

		# Report success
		return True

	def var_info(self):

		"""
		Give a list of dictionaries with variable descriptions

		Returns:
		A list of (name, description) tuples
		"""

		return item.item.var_info(self) + \
			generic_response.generic_response.var_info(self)

class qtport_reader(port_reader, qtplugin.qtplugin):

	"""GUI controls for port_reader plug-in"""

	def __init__(self, name, experiment, string=None):

		"""
		Constructor

		Arguments:
		name -- item name
		experiment -- experiment instance

		Keyword arguments:
		string -- definition string (default=None)
		"""

		# Pass the word on to the parents
		port_reader.__init__(self, name, experiment, string)
		qtplugin.qtplugin.__init__(self, __file__)

	def init_edit_widget(self):

		"""Initialize GUI controls"""

		# Lock the widget until we're doing creating it
		self.lock = True

		# Pass the word on to the parent
		qtplugin.qtplugin.init_edit_widget(self, False)

		# Create the controls
		#
		# A number of convenience functions are available which
		# automatically create controls, which are also automatically
		# updated and applied. If you set the varname to None, the
		# controls will be created, but not automatically updated
		# and applied.
		#
		# qtplugin.add_combobox_control(varname, label, list_of_options)
		# - creates a QComboBox
		# qtplugin.add_line_edit_control(varname, label)
		# - creates a QLineEdit
		# qtplugin.add_spinbox_control(varname, label, min, max, suffix = suffix, prefix = prefix)

		self.add_combobox_control("dummy", \
			"Dummy mode (use keyboard instead)", ["yes", "no"], \
			tooltip="Enable dummy mode for testing purposes")
		self.add_spinbox_control("port", "Port number", 0, 1024, tooltip= \
			"Expecting a valid port number")
		self.add_spinbox_control("resting_value", "Resting state value", 0, \
			255, tooltip= \
			"A value that is read from the port when there is no input")
		self.add_line_edit_control("correct_response", "Correct response", \
			tooltip="Expecting response values (0 .. 255)")
		self.add_line_edit_control("timeout", "Timeout", tooltip= \
			"Expecting a value in milliseconds or 'infinite'", default= \
			"infinite")

		# Add a stretch to the edit_vbox, so that the controls do not
		# stretch to the bottom of the window.
		self.edit_vbox.addStretch()

		# Unlock
		self.lock = True

	def apply_edit_changes(self):

		"""Apply GUI controls"""

		# Abort if the parent reports failure of if the controls are locked
		if not qtplugin.qtplugin.apply_edit_changes(self) or self.lock:
			return False

		# Refresh the main window, so that changes become visible everywhere
		self.experiment.main_window.refresh(self.name)

		# Report success
		return True

	def edit_widget(self):

		"""Update GUI controls"""

		# Lock the controls, otherwise a recursive loop might aris
		# in which updating the controls causes the variables to be
		# updated, which causes the controls to be updated, etc...
		self.lock = True

		# Let the parent handle everything
		qtplugin.qtplugin.edit_widget(self)

		# Unlock
		self.lock = False

		# Return the _edit_widget
		return self._edit_widget



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *

class items_adapter(object):

	"""
	desc:
		Gives an editable view on (item_name, start_time, end_time, cond) tuple
		(schedule) used by coroutines that funtions like an (item_name, cond)
		tuple as used by sequences.
	"""

	def __init__(self, schedule):

		self.schedule = schedule

	def __delitem__(self, index):

		del self.schedule[index]

	def __getitem__(self, index):

		return self.schedule[index][0], self.schedule[index][3]

	def __setitem__(self, index, val):

		self.schedule[index] = (val[0],) + self.schedule[index][1:3] + (val[1],)

	def __iter__(self):

		for item_name, start_time, end_time, cond in self.schedule:
			yield item_name, cond

	def __len__(self):

		return len(self.schedule)

	def insert(self, index, item):

		self.schedule.insert(index, ( (item[0],) + (0, 0) + (item[1],) ) )

	def append(self, item):

		self.schedule.append( ( (item[0],) + (0, 0) + (item[1],) ) )



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from qtpy import QtCore, QtGui, QtWidgets
from libopensesame import debug
from libqtopensesame.misc.base_subcomponent import base_subcomponent
from libqtopensesame.misc import drag_and_drop
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'toolbar_items_item', category=u'core')

class toolbar_items_item(base_subcomponent, QtWidgets.QLabel):

	"""
	desc:
		A draggable toolbar icon.
	"""

	def __init__(self, parent, item, pixmap=None):

		"""
		desc:
			Constructor.

		arguments:
			parent:
				desc:	The parent.
				type:	QWidget
			item:
				desc:	An item.
				type:	qtitem
			pixmap:
				desc:	A pixmap for the icon.
				type:	QPixmap
		"""

		super(toolbar_items_item, self).__init__(parent)
		self.setup(parent)
		self.item = item
		if pixmap is None:
			self.pixmap = self.theme.qpixmap(item)
		else:
			self.pixmap = pixmap
		# self.setMargin(6)
		self.setToolTip(_("Drag this <b>%s</b> item to the intended location in the overview area or into the item list of a sequence tab") \
			% self.item)
		self.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
		self.setPixmap(self.pixmap)

	def mousePressEvent(self, e):

		"""
		desc:
			Initiates a drag event after a mouse press.

		arguments:
			e:
				desc:	A mouse event.
				type:	QMouseEvent
		"""

		if e.buttons() != QtCore.Qt.LeftButton:
			return
		name = u'new_%s' % self.item
		data = {
			u'type'				: u'item-snippet',
			u'main-item-name'	: name,
			u'items'			: [{
				u'item-type'	: self.item,
				u'item-name'	: name,
				u'script'		: u''
				}]
			}
		self.item_toolbar.collapse()
		drag_and_drop.send(self, data)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import math
from openexp._canvas._polygon.polygon import Polygon


class Arrow(Polygon):

	def __init__(self, canvas, sx, sy, ex, ey, body_length=0.8, body_width=.5,
		head_width=30, **properties):

		self._create_property('sx')
		self._create_property('sy')
		self._create_property('ex')
		self._create_property('ey')
		self._create_property('body_length')
		self._create_property('body_width')
		self._create_property('head_width')
		properties = properties.copy()
		properties.update({
			'sx': sx,
			'sy': sy,
			'ex': ex,
			'ey': ey,
			'body_length': body_length,
			'body_width': body_width,
			'head_width': head_width,
		})
		Polygon.__init__(
			self,
			canvas,
			self._shape(sx, sy, ex, ey, body_length, body_width, head_width),
			**properties
		)

	@staticmethod
	def _shape(sx, sy, ex, ey, body_length, body_width, head_width):

		"""
		returns:
			Returns a list of (x, y) tuples that specify an arrow shape.
		"""

		# length
		d = math.sqrt((ey-sy)**2 + (sx-ex)**2)
		# direction
		angle = math.atan2(ey-sy, ex-sx)
		_head_width = (1-body_width)/2.0
		body_width = body_width/2.0
		# calculate coordinates
		p4 = (ex, ey)
		p1 = (sx +body_width * head_width * math.cos(angle - math.pi/2),
			sy + body_width * head_width * math.sin(angle - math.pi/2))
		p2 = (p1[0] + body_length*math.cos(angle) * d,
			p1[1] + body_length * math.sin(angle) * d)
		p3 = (p2[0]+_head_width * head_width * math.cos(angle-math.pi/2),
			p2[1] + _head_width * head_width * math.sin(angle-math.pi/2))
		p7 = (sx + body_width * head_width*math.cos(angle + math.pi/2),
			sy + body_width * head_width*math.sin(angle + math.pi/2))
		p6 = (p7[0] + body_length * math.cos(angle) * d,
			p7[1] + body_length * math.sin(angle) * d)
		p5 = (p6[0]+_head_width * head_width * math.cos(angle+math.pi/2),
			p6[1]+_head_width * head_width * math.sin(angle+math.pi/2))
		return [p1, p2, p3, p4, p5, p6, p7]



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from openexp import backend
from libopensesame.var_store import var_store
from libopensesame.item_store import item_store
from libopensesame.response_store import response_store
from libopensesame.file_pool_store import file_pool_store
from libopensesame.syntax import syntax
from libopensesame.exceptions import osexception
from libopensesame import misc, item, debug, metadata
from libopensesame.item_stack import item_stack_singleton
from libopensesame.oslogging import oslogger
from libopensesame.py3compat import *
import os
import pickle
import time
import warnings
import gc


# In Python 2, sending old-style classes across the pipeline causes trouble. We
# need to define the classobj type here. This is only applicable to Python 2,
# because Python 3 had only new-style classes.
class OldStyle: pass
classobj = type(OldStyle)


class experiment(item.item):

	"""
	desc:
		A special item that controls the flow of the experiment.
	"""

	description = u'The main experiment item'

	def __init__(self, name=u'experiment', string=None, pool_folder=None,
		experiment_path=None, fullscreen=False, auto_response=False,
		logfile=u'defaultlog.csv', subject_nr=0, workspace=None, resources={},
		heartbeat_interval=1):

		"""
		desc:
			Constructor. The experiment is created automatically be OpenSesame
			and you will generally not need to create it yourself.

		keywords:
			name:
				desc:	The name of the experiment.
				type:	[str, unicode]
			string:
				desc:	A string containing the experiment definition, the
						name of an OpenSesame experiment file, or `None` to
						create a blank experiment.
				type:	[str, unicode, NoneType]
			pool_folder:
				desc:	A specific folder to be used for the file pool, or
						`None` to use a new temporary folder.
				type:	[str, unicode, NoneType]
			experiment_path:
				desc:	The path of the experiment file. This is the folder that
						the experiment is in, not the path to the experiment
						file.
				type:	str
			fullscreen:
				desc:	Indicates whether the experiment should be executed in
						fullscreen.
				type:	bool
			auto_response:
				desc:	Indicates whether auto-response mode should be enabled.
				type:	bool
			logfile:
				desc:	The logfile path.
				type:	[unicode, str]
			subject_nr:
				desc:	The subject number.
				type:	int
			workspace:
				desc:	A `python_workspace` object to be used for executing
						custom Python code, or `None` to create a new workspace.
				type:	[python_workspace, NoneType]
			resources:
				desc:	A dictionary with names as keys and paths as values.
						This serves as a look-up table for resources.
				type:	dict
			heartbeat_interval:
				desc:	A heartbeat interval in seconds, or <= 0 to disable
						heartbeats.
				type:	[int, float]
		"""

		# Make sure the logger is started
		if not oslogger.started:
			oslogger.start()
			oslogger.debug('starting logger in experiment init')
		self.var = var_store(self)
		self.pool = file_pool_store(self, folder=pool_folder)
		self._responses = response_store(self)
		# The _syntax and items objects may already have been created by
		# libqtopensesame.experiment.
		if not hasattr(self, u'_syntax'):
			self._syntax = syntax(self)
		if not hasattr(self, u'items'):
			self.items = item_store(self)
		if workspace is None:
			from libopensesame.python_workspace import python_workspace
			self._python_workspace = python_workspace(self)
		else:
			self._python_workspace = workspace
		self.running = False
		self.auto_response = auto_response
		self.heartbeat_interval = heartbeat_interval
		self.plugin_folder = u'plugins'
		self._start_response_interval = None
		self.cleanup_functions = []
		self.restart = False
		self.resources = resources
		self.paused = False
		self.output_channel = None
		self.reset()

		# Logfile parameters
		self._log = None
		self.logfile = logfile
		# A list of data files, which may include more than just the OpenSesame
		# logfile, for example if data is also recorded using some other method.
		self.data_files = []

		# This is some duplication of the option parser in qtopensesame,
		# but nevertheless keep it so we don't need qtopensesame
		self.debug = debug.enabled
		if string is not None:
			string = self.open(string)
		item.item.__init__(self, name, self, string)
		# Default subject info
		self.set_subject(subject_nr)
		# Fullscreen needs to be set after the experiment has been parsed from
		# script, otherwise it will be overridden by the script.
		self.var.fullscreen = u'yes' if fullscreen else u'no'
		# Restore experiment path, which is either the full path (including
		# filename), only the folder, or None.
		if experiment_path is not None:
			if os.path.isfile(experiment_path):
				self.var.experiment_path, self.var.experiment_file = \
					self.experiment_path, self.experiment_file = \
						os.path.split(experiment_path)
			else:
				self.var.experiment_path = self.experiment_path \
					= experiment_path
		if not py3 and backend.backend_guess(self, 'sampler') == u'psycho':
			self.var.sampler_backend = 'legacy'
			oslogger.warning(
				u'The psycho sampler backend is not compatible with Python 2. '
				u'Falling back to legacy sampler backend'
			)


	@property
	def pool_folder(self):

		"""Deprecated."""

		warnings.warn(
			u'experiment.pool_folder is deprecated. Use file_pool_store instead.',
			DeprecationWarning)
		return self.pool.folder()

	@property
	def fallback_pool_folder(self):

		"""Deprecated."""

		warnings.warn(
			u'experiment.fallback_pool_folder is deprecated. Use file_pool_store instead.',
			DeprecationWarning)
		return self.pool.fallback_folder()

	@property
	def default_title(self):
		return u'New experiment'

	def get_file(self, path):

		"""Deprecated."""

		warnings.warn(
			u'experiment.get_file() is deprecated. Use file_pool_store instead.',
			DeprecationWarning)
		return self.pool[path]

	def reset(self):

		"""See item."""

		# Set default variables
		self.var.start = u'experiment'
		self.var.title = self.default_title
		self.var.round_decimals = 2
		self.var.form_clicks = u'no'
		self.var.disable_garbage_collection = u'yes'
		# In version 2.9.X and before, the sketchpad used 0,0 for the screen
		# center, whereas scripting items used 0,0 for the top-left. By setting
		# uniform_coordinates to 'yes', 0,0 is used for the center in all cases.
		self.var.uniform_coordinates = u'no'
		# Sound parameters
		self.var.sound_freq = 48000
		self.var.sound_sample_size = -16 # Negative values mean signed
		self.var.sound_channels = 2
		self.var.sound_buf_size = 1024
		# Default backend
		self.var.canvas_backend = u'xpyriment'
		# Display parameters
		self.var.width = 1024
		self.var.height = 768
		self.var.background = u'black'
		self.var.foreground = u'white'
		# Font parameters
		self.var.font_size = 18
		self.var.font_family = u'mono'
		self.var.font_italic = u'no'
		self.var.font_bold = u'no'
		self.var.font_underline = u'no'

	def file_in_pool(self, path):

		"""Deprecated."""

		warnings.warn(
			u'experiment.file_in_pool() is deprecated. Use file_pool_store instead.',
			DeprecationWarning)
		return path in self.pool

	def module_container(self):

		"""Specify the module that contains the item modules"""

		return u'libopensesame'

	def item_prefix(self):

		"""
		A prefix for the plug-in classes, so that [prefix][plugin] class is used
		instead of the [plugin] class.
		"""

		return u''

	def set_subject(self, nr):

		"""
		desc:
			Sets the subject number and parity (even/ odd). This function is
			called automatically when an experiment is started, so you do not
			generally need to call it yourself.

		arguments:
			nr:
				desc:	The subject nr.
				type:	int

		example: |
			exp.set_subject(1)
			print('Subject nr = %d' % exp.get('subject_nr'))
			print('Subject parity = %s' % exp.get('subject_parity'))
		"""

		# Set the subject nr and parity
		self.var.subject_nr = nr
		if nr % 2 == 0:
			self.var.subject_parity = u'even'
		else:
			self.var.subject_parity = u'odd'

	def read_definition(self, s):

		"""
		Extracts a the definition of a single item from the string.

		Arguments:
		s	--	The definition string.

		Returns:
		A (str, str) tuple with the full string minus the definition string
		and the definition string.
		"""

		# Read the string until the end of the definition
		def_str = u''
		line = next(s, None)
		if line is None:
			return None, u''
		get_next = False
		while True:
			if len(line) > 0:
				if line[0] != u'\t':
					break
				else:
					def_str += line + u'\n'
			line = next(s, None)
			if line is None:
				break
		return line, def_str

	def from_string(self, string):

		"""
		desc:
			Reads the entire experiment from a string.

		arguments:
			string:	The definition string.
		"""

		self.var.clear(preserve=[u'experiment_path', u'experiment_file'])
		self.reset()
		self.comments = []
		oslogger.debug(u"building experiment")
		if string is None:
			return
		self.front_matter, string = self._syntax.parse_front_matter(string)
		if self.experiment.front_matter[u'API'].version[0] < 2:
			# Old experiment scripts were saved in ASCII, and require decoding
			# of U+XXXX unicode characters.
			string = self.syntax.from_ascii(string)
		s = iter(string.split(u'\n'));
		line = next(s, None)
		while line is not None:
			get_next = True
			try:
				l = self.syntax.split(line)
			except ValueError as e:
				raise osexception(
					u'Failed to parse script. Maybe it contains '
					u'illegal characters or unclosed quotes?',
					exception=e
				)
			if l:
				self.parse_variable(line)
				# Parse definitions
				if l[0] == u"define":
					if len(l) != 3:
						raise osexception(
							u'Failed to parse definition',
							line=line
						)
					item_type = l[1]
					item_name = self.syntax.sanitize(l[2])
					line, def_str = self.read_definition(s)
					get_next = False
					self.items.new(
						_type=item_type,
						name=item_name,
						script=def_str,
						allow_rename=False
					)
			# Advance to next line
			if get_next:
				line = next(s, None)

	def transmit_workspace(self, **extra):

		"""
		desc:
			Sends the current workspace through the output channel. If there is
			no output channel, this function does nothing.

		keyword-dict:
			extra:	Any extra items in the workspace dict to be sent.
		"""

		if self.output_channel is None:
			return
		d = self.python_workspace._globals.copy()
		d.update(extra)
		for key, value in d.copy().items():
			if self._is_oldstyle(value):
				del d[key]
				continue
			try:
				pickle.dumps(value)
			except:
				del d[key]
		self.output_channel.put(d)

	def _is_oldstyle(self, obj):

		"""
		desc:
			Checks if an object is an old-style class or an instance of an
			old-style class. This is important, because these objects cannot be
			pickled properly, and the workspace transmission chokes on them.
		"""

		if py3:
			return False
		# old-style class
		if type(obj) == classobj:
			return True
		# instance of old-style class
		if hasattr(obj, u'__class__') and type(obj.__class__) == classobj:
			return True
		return False

	def set_output_channel(self, output_channel):

		"""
		desc:
			Sets the output channel, which is used to communicate the workspace
			between the experiment and the launch process (typically the GUI).

		arguments:
			output_channel:
				desc: 	The output object, which must support a `put` method.
		"""

		if not hasattr(output_channel, u'put'):
			raise osexception(u'Invalid output_channel: %s' % output_channel)
		self.output_channel = output_channel

	def run(self):

		"""Runs the experiment."""

		# Save the date and time, and the version of OpenSesame
		self.var.datetime = safe_decode(
			time.strftime(u'%c'),
			enc=self.encoding,
			errors=u'ignore'
		)
		self.var.opensesame_version = metadata.__version__
		self.var.opensesame_codename = metadata.codename
		self.running = True
		self.init_random()
		self.init_display()
		self.init_clock()
		self.init_sound()
		self.init_log()
		self.python_workspace.init_globals()
		self.reset_feedback()
		self.init_heartbeat()
		if self.var.uniform_coordinates != u'yes':
			oslogger.warning(
				u'Uniform coordinates are disabled. This is deprecated, and '
				u'will be removed in the future. Please enable uniform '
				u'coordinates in the script of your experiment.'
			)
		oslogger.info(u"experiment started")
		if self.var.start in self.items:
			item_stack_singleton.clear()
			if self.var.disable_garbage_collection == u'yes':
				oslogger.info('disabling garbage collection')
				gc.disable()
			self.items.execute(self.var.start)
		else:
			raise osexception(
				"Could not find item '%s', which is the entry point of the experiment"
				% self.var.start
			)
		oslogger.info(u"experiment finished")
		self.end()

	def pause(self):

		"""
		desc:
			Pauses the experiment, sends the Python workspace to the GUI, and
			waits for the GUI to send a resume signal. This requires an output
			channel.
		"""

		if self.paused:
			return

		from openexp.canvas import Canvas
		from openexp.canvas_elements import Text
		from openexp.keyboard import Keyboard

		self.paused = True
		self.transmit_workspace(__pause__=True)
		pause_canvas = Canvas(self)
		pause_canvas['msg'] = Text(
			u'The experiment has been paused<br /><br />'
			u'Press spacebar to resume<br />'
			u'Press Q to quit'
		).construct(pause_canvas)
		pause_keyboard = Keyboard(
			self,
			keylist=[u'space', u'q', u'Q'],
			timeout=0
		)
		pause_keyboard.show_virtual_keyboard()
		pause_canvas.show()
		try:
			while True:
				key, _time = pause_keyboard.get_key()
				if key == u'q' or key == u'Q':
					pause_keyboard.show_virtual_keyboard(False)
					raise osexception(
						u'The experiment was aborted',
						user_triggered=True
					)
				if key == u'space':
					break
				time.sleep(.25)
		finally:
			self.paused = False
			self.transmit_workspace(__pause__=False)
		pause_keyboard.show_virtual_keyboard(False)
		pause_canvas['msg'].text = u'Resuming ...'
		pause_canvas.show()

	def cleanup(self):

		"""Calls all the cleanup functions."""

		while len(self.cleanup_functions) > 0:
			func = self.cleanup_functions.pop()
			oslogger.debug(u"calling cleanup function")
			func()

	def end(self):

		"""Nicely ends the experiment."""

		from openexp import sampler, canvas
		self.running = False
		try:
			self._log.close()
		except AttributeError:
			oslogger.error('missing or invalid log object')
		sampler.close_sound(self)
		canvas.close_display(self)
		self.cleanup()
		if not gc.isenabled():
			oslogger.info('enabling garbage collection')
			gc.enable()
		self.transmit_workspace(__finished__=True)

	def to_string(self):

		"""
		Encodes the experiment into a string.

		Returns:
		A Unicode definition string for the experiment.
		"""

		s = self._syntax.generate_front_matter()
		for var in self.var:
			s += self.variable_to_string(var)
		s += u'\n'
		for _item in sorted(self.items):
			s += self.items[_item].to_string() + u'\n'
		return s

	def resource(self, name):

		"""
		Retrieves a file from the resources folder.

		Arguments:
		name	--	The file name.

		Returns:
		A Unicode string with the full path to the file in the resources
		folder.
		"""

		name = safe_decode(name)
		if self is not None:
			if name in self.resources:
				return self.resources[name]
			if os.path.exists(self.pool[name]):
				return self.pool[name]
		path = misc.resource(name)
		if path is None:
			raise FileNotFoundError(
				u"The resource '%s' could not be found in libopensesame.experiment.resource()"
				% name
			)
		return path

	def save(self, path, overwrite=False, update_path=True):

		"""
		desc:
			Saves the experiment to file.

		arguments:
			path:
				desc:	The target file to save to.
				type:	[str, unicode]

		keywords:
			overwrite:
				desc:	Indicates if existing files should be overwritten.
				type:	bool
			update_path:
				desc:	Indicates if the experiment_path attribute should be
						updated.
				type:	bool

		returns:
			desc:	The path on successful saving or False otherwise.
			type:	[unicode, bool]
		"""

		path = safe_decode(path, enc=self.encoding)
		if os.path.exists(path) and not overwrite:
			return False
		from libopensesame.osexpfile import osexpwriter
		w = osexpwriter(self, path)
		if update_path:
			self.experiment_path = w.experiment_path
		return path

	def open(self, src):

		"""
		desc: |
			Opens an experiment. The source can be any of the following:

			- An OpenSesame script (ie. not a file)
			- The full path to a .osexp experiment file.

		arguments:
			src:
				desc:	The source.
				type:	str

		returns:
			desc:	An OpenSesame script.
			type:	str
		"""

		from libopensesame.osexpfile import osexpreader
		f = osexpreader(self, src)
		self.experiment_path = f.experiment_path
		return f.script

	def reset_feedback(self):

		"""Resets the feedback variables (acc, avg_rt, etc.)."""

		self.responses.reset_feedback()

	def var_info(self):

		"""
		Returns a list of (name, value) tuples with variable descriptions
		for the main experiment.

		Returns:
		A list of tuples.
		"""

		l = []
		for var in self.var:
			l.append( (var, self.var.get(var, _eval=False)) )
		return l

	def init_heartbeat(self):

		"""
		desc:
			Initializes heartbeat.
		"""

		if self.heartbeat_interval <= 0 or self.var.fullscreen == u'yes' or \
			self.output_channel is None:
			self.heartbeat = None
			return
		from libopensesame.heartbeat import heartbeat
		self.heartbeat = heartbeat(self, interval=self.heartbeat_interval)
		self.heartbeat.start()

	def init_random(self):

		"""
		desc:
			Initializes the random number generators. For some reason, the numpy
			random seed is not re-initialized when the experiment is started
			again with the multiprocess runner, resulting in identical random
			runs. The standard random module doesn't suffer from this problem.
			But to be on the safe side, we now explicitly re-initialize the
			random seed.

			See also:

			- <http://forum.cogsci.nl/index.php?p=/discussion/1441/>
		"""

		import random
		random.seed()
		try:
			# Don't assume that numpy is available
			import numpy
			numpy.random.seed()
		except:
			pass

	def init_sound(self):

		"""Intializes the sound backend."""

		from openexp import sampler
		sampler.init_sound(self)

	def init_display(self):

		"""Initializes the canvas backend."""

		from openexp import canvas
		canvas.init_display(self)
		self.python_workspace[u'win'] = self.window

	def init_clock(self):

		"""Initializes the clock backend."""

		from openexp.clock import clock
		self._clock = clock(self)

	def init_log(self):

		"""Initializes the log backend."""

		from openexp.log import log
		self._log = log(self, self.logfile)

def clean_up(verbose=False, keep=[]):

	warnings.warn(u'libopensesame.experiment.clean_up() is deprecated',
		DeprecationWarning)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *

from libopensesame.exceptions import osexception
from libopensesame import item
import gc
import openexp.keyboard

class sequence(item.item):

	"""The sequence item"""

	description = u'Runs a number of items in sequence'

	def reset(self):

		"""See item."""

		self.items = []
		self.var.flush_keyboard = u'yes'

	def run(self):

		"""Runs the sequence."""

		self.set_item_onset()
		# Optionally flush the responses to catch escape presses
		if self._keyboard is not None:
			self._keyboard.flush()
		for _item, cond in self._items:
			self.python_workspace[u'self'] = self
			if self.python_workspace._eval(cond):
				self.experiment.items.run(_item)
		if not gc.isenabled():
			gc.collect()

	def set_validator(self):

		"""
		desc:
			See qtitem.
		"""

		self.validator = sequence

	def parse_run(self, i):

		"""
		Parses a run line from the definition script.

		Arguments:
		i 		-- 	A list of words, corresponding to a single script line.

		Returns:
		An (item_name, conditional) tuple.
		"""

		name = i[1]
		cond = u'always'
		if len(i) > 2:
			cond = i[2]
		return name, cond

	def from_string(self, string):

		"""
		Parses a definition string.

		Arguments:
		string 	--	A definition string.
		"""

		self.var.clear()
		self.comments = []
		self.reset()
		if string is None:
			return
		for s in string.split(u'\n'):
			self.parse_variable(s)
			cmd, arglist, kwdict = self.syntax.parse_cmd(s)
			if cmd != u'run' or not len(arglist):
				continue
			_item = arglist[0]
			if len(arglist) == 1:
				cond = u'always'
			else:
				cond = arglist[1]
			self.items.append((_item, cond))

	def prepare(self):

		"""Prepares the sequence."""

		item.item.prepare(self)
		if self.var.flush_keyboard == u'yes':
			# Create a keyboard to flush responses at the start of the run phase
			self._keyboard = openexp.keyboard.keyboard(self.experiment)
		else:
			self._keyboard = None
		self._items = []
		for _item, cond in self.items:
			if _item not in self.experiment.items:
				raise osexception( \
					u"Could not find item '%s', which is called by sequence item '%s'" \
					% (_item, self.name))
			self.experiment.items.prepare(_item)
			self._items.append( (_item, self.syntax.compile_cond(cond)) )

	def to_string(self):

		"""
		Encodes the sequence as a definition string.

		Returns:
		A definition string.
		"""

		s = item.item.to_string(self, self.item_type)
		for _item, cond in self.items:
			s += u'\t' + self.syntax.create_cmd(u'run', [_item, cond]) + u'\n'
		return s



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception, AbortCoroutines

class base_task(object):

	"""
	desc:
		A task controls the coroutine for one item.
	"""

	ALIVE = 0
	DEAD = 1
	ABORT = 2

	def __init__(self, coroutines, start_time, end_time, abort_on_end=False):

		"""
		desc:
			Constructor.

		arguments:
			start_time:
				desc:	The start-time of the coroutine.
				type:	int
			end_time:
				desc:	The end-time of the coroutine.
				type:	int

		keywords:
			abort_on_end:
				desc:	Indicates whetehr an ABORT signal should be given when
						the task ends.
				type:	bool`
		"""

		if not (isinstance(start_time, (int, float)) and start_time >= 0) or \
			not (isinstance(end_time, (int, float)) and end_time >= start_time):
			raise osexception(
				(u'Start (now: %s) and end (now: %s) time need to be '
				u'non-negative numeric values and end time needs to be equal to '
				u' or higher than start time') % (start_time, end_time))
		self.start_time = start_time
		self.end_time = end_time
		self.coroutines = coroutines
		self._abort_on_end = abort_on_end

	def started(self, dt):

		"""
		arguments:
			dt:
				desc:	The current time relative to onset of the coroutines.
				type:	int

		returns:
			desc:	True if the current item is started, False otherwise.
			type:	bool
		"""

		return dt >= self.start_time

	def stopped(self, dt):

		"""
		desc:
			Checks whether an item is stopped, and sends the stop signal to the
			item if it should be stopped.

		arguments:
			dt:
				desc:	The current time relative to onset of the coroutines.
				type:	int

		returns:
			desc:	True if the current item is started, False otherwise.
			type:	bool
		"""

		if dt < self.end_time:
			return False
		try:
			self.coroutine.send(False)
		except StopIteration:
			self.coroutines.event('stopped %s' % self.coroutine)
		return True

	def kill(self):

		"""
		desc:
			Sends the stop signal to an item.
		"""

		try:
			self.coroutine.send(False)
		except StopIteration:
			self.coroutines.event('killed %s' % self.coroutine)

	def launch(self):

		"""
		desc:
			Launches an item. All items are launched at coroutine start.
		"""

		raise NotImplementedError()

	def step(self):

		"""
		desc:
			Lets the item yield one cycle.
		"""

		try:
			if self.coroutine.send(True) is False:
				return self.ABORT
		except AbortCoroutines:
			return self.ABORT
		except StopIteration:
			self.coroutines.event('died %s' % self.coroutine)
			return self.ABORT if self._abort_on_end else self.DEAD
		return self.ALIVE



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *

from libqtopensesame.extensions import base_extension

class plugin_manager(base_extension):

	"""
	desc:
		Manages plug-ins and extensions
	"""

	def event_startup(self):

		self._widget = None

	def activate(self):

		"""
		desc:
			Open the plug-in manager tab, or switch to it if already open.
		"""

		self.tabwidget.add(self.widget(), self.icon(), self.label())

	def widget(self):

		"""
		returns:
			desc:	The widget to be used in the tab widget.
			type:	QWidget
		"""

		if self._widget is None:
			self.set_busy()
			from plugin_manager_widget import plugin_manager_widget
			self._widget = plugin_manager_widget(self.main_window)
			self.set_busy(False)
		return self._widget



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.misc.base_subcomponent import base_subcomponent
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'sketchpad', category=u'item')

class sketchpad_element_button(base_subcomponent, QtWidgets.QPushButton):

	"""
	desc:
		A element-tool button for the sketchpad widget.
	"""

	def __init__(self, sketchpad_widget, element):

		"""
		desc:
			Constructor.

		arguments:
			sketchpad_widget:
				type:	sketchpad-widget
			element:
				type:	base_element
		"""

		super(sketchpad_element_button, self).__init__(sketchpad_widget)
		self.element = element
		self.sketchpad_widget = sketchpad_widget
		self.setup(sketchpad_widget)
		self.setFlat(True)
		self.setCheckable(True)
		self.setIcon(self.theme.qicon(u'os-%s' % self.name))
		self.setToolTip(_(u'Draw %s element') % self.name)
		self.clicked.connect(self.select)

	def select(self, checked):

		"""
		desc:
			Select this element tool.
		"""

		self.sketchpad_widget.unselect_all_tools()
		self.setChecked(True)
		self.sketchpad_widget.select_element_tool(self.element)

	@property
	def name(self):
		return self.element.__name__



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from qtpy import QtCore, QtGui, QtWidgets
from libqtopensesame.widgets.base_widget import base_widget
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'font_widget', category=u'core')


class font_widget_base(base_widget):

	"""A font selection widget"""

	max_size = 64
	font_list = [
		u'mono', u'sans', u'serif', u'arabic',
		u'chinese-japanese-korean', u'hebrew', u'hindi',
		_(u'other ...')
	]

	font_changed = QtCore.Signal(['QString', int, bool, bool])

	def _apply(self):

		"""Applies the controls."""

		self.family = self.ui.combobox_family.currentText()
		self.size = self.ui.spinbox_size.value()
		self.italic = self.ui.checkbox_italic.isChecked()
		self.bold = self.ui.checkbox_bold.isChecked()
		self.ui.label_example.setFont(self.get_font())
		self.font_changed.emit(
			self.family,
			self.size,
			self.italic,
			self.bold
		)

	def apply_family(self):

		"""
		Applies the controls and optionally presents a full font selection
		dialog if the user has selected 'other ...'
		"""

		if self.ui.combobox_family.currentText() == _(
			u'other ...',
			context=u'font_widget'
		):
			font, ok = QtWidgets.QFontDialog.getFont(
				self.get_font(),
				parent=self._parent
			)
			if ok:
				self.family = str(font.family())
			else:
				self.family = self.experiment.var.get(
					u'font_family',
					_eval=False
				)
			self.update_family_combobox()
		self._apply()

	def get_font(self):

		"""
		returns:
			type: QFont
		"""

		font = QtGui.QFont(
			self.family,
			italic=self.italic,
			weight=QtGui.QFont.Bold if self.bold else QtGui.QFont.Normal
		)
		font.setPixelSize(min(self.max_size, self.size))
		return font

	def set_font(self, family=None, italic=None, bold=None, size=None):

		"""
		desc:
			Sets the widget to the specified font. For keywords, see
			initialize().
		"""

		self.ui.combobox_family.activated.disconnect()
		self.ui.checkbox_italic.clicked.disconnect()
		self.ui.checkbox_bold.clicked.disconnect()
		self.ui.spinbox_size.editingFinished.disconnect()
		self.initialize(
			family=family,
			italic=italic,
			bold=bold,
			size=size,
			parent=self._parent
		)

	def initialize(
		self,
		experiment=None,
		family=None,
		italic=None,
		bold=None,
		size=None,
		parent=None
	):

		"""
		desc:
			Initializes the widget.

		arguments:
			experiment:		The experiment.

		keywords:
			family:	The font family or None to use experiment default.
			italic:	The font italic state or None to use experiment
					default.
			bold:	font bold state or None to use experiment default.
			size:	The font size or None to use experiment default.
			parent:	A parent QWidget.
		"""

		self._parent = parent if parent is not None else self.main_window
		if experiment is None:
			experiment = self.experiment
		self.family = experiment.var.get(u'font_family', _eval=False) \
			if family is None else family
		self.italic = experiment.var.get(u'font_italic', _eval=False) == u'yes' \
			if italic is None else italic
		self.bold = experiment.var.get(u'font_bold', _eval=False) == u'yes' \
			if bold is None else bold
		self.size = experiment.var.get(u'font_size', _eval=False) \
			if size is None else size
		if not isinstance(self.size, int) or self.size <= 0:
			self.size = 18
		if self.ui.combobox_family.findText(self.family) < 0:
			self.ui.combobox_family.addItem(self.family)
		self.ui.combobox_family.setCurrentIndex(
			self.ui.combobox_family.findText(self.family))
		self.ui.checkbox_italic.setChecked(self.italic)
		self.ui.checkbox_bold.setChecked(self.bold)
		self.ui.spinbox_size.setValue(self.size)
		self._apply()
		self.ui.combobox_family.activated.connect(self.apply_family)
		self.ui.checkbox_italic.clicked.connect(self._apply)
		self.ui.checkbox_bold.clicked.connect(self._apply)
		self.ui.spinbox_size.editingFinished.connect(self._apply)

	def update_family_combobox(self):

		"""Updates the family combobox to include a custom font."""

		self.ui.combobox_family.activated.disconnect()
		self.ui.combobox_family.clear()
		l = self.font_list[:]
		if self.family not in l:
			l += [self.family]
		self.ui.combobox_family.insertItems(0, l)
		self.ui.combobox_family.setCurrentIndex(
			self.ui.combobox_family.findText(self.family))
		self.ui.combobox_family.activated.connect(self.apply_family)


class font_widget_horizontal(font_widget_base):

	def __init__(self, main_window):

		"""
		Constructor.

		Arguments:
		experiment	--	The experiment.

		Keywords arguments:
		parent		--	The parent QWidget. (default=None)
		"""

		super(font_widget_horizontal, self).__init__(
			main_window,
			ui=u'widgets.font_widget_horizontal'
		)


class font_widget(font_widget_base):

	def __init__(self, main_window):

		"""
		Constructor.

		Arguments:
		experiment	--	The experiment.

		Keywords arguments:
		parent		--	The parent QWidget. (default=None)
		"""

		super(font_widget, self).__init__(
			main_window,
			ui=u'widgets.font_widget'
		)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import sys
import logging
import logging.handlers


MAXBYTES = 10 * 1024 ** 2
BACKUPCOUNT = 1


class OSLogger(object):

	"""
	desc:
		Implements application logging through the Python logging module.
	"""

	def start(self, name=u'default'):

		"""
		desc:
			When the module is imported, an uninitialized singleton instance of
			OSLogger is created. OSLogger.start() is then called to Initialize
			the actual logger.
		"""

		self._name = name
		self._formatter = logging.Formatter(
			u'[%(asctime)s:%(module)s:%(lineno)s:%(levelname)s] %(message)s'
		)
		self._logger = logging.getLogger(name)
		self._logger.propagate = False
		self.add_handler(self.StreamHandler(sys.stdout))
		if self.debug_mode:
			self._logger.setLevel(logging.DEBUG)
			try:
				self.add_handler(self.RotatingFileHandler())
			except PermissionError:
				self.error(u'failed to set RotatingFileHandler')
		else:
			self._logger.setLevel(logging.INFO)

	def RotatingFileHandler(self, path=None, level=logging.DEBUG):

		"""
		desc:
			A factory that returns an instance of a RotatingFileHandler.

		keywords:
			path:   A logfile path or `None` to use the default path.
			level:  A logging level.

		returns:
			A RotatingFileHandler.
		"""

		if path is None:
			from libopensesame import misc
			path = os.path.join(
				misc.home_folder(),
				u'.opensesame',
				u'debug.log'
			)
		self.info('debug info in %s' % path)
		h = logging.handlers.RotatingFileHandler(
			path,
			maxBytes=MAXBYTES,
			backupCount=BACKUPCOUNT,
			encoding=u'utf-8'
		)
		h.setLevel(level)
		h.setFormatter(self._formatter)
		return h

	def StreamHandler(self, stream, level=logging.INFO):

		"""
		desc:
			A factory that returns an instance of a StreamHandler.

		keywords:
			stream: A file-like object.
			level:  A logging level.

		returns:
			A StreamHandler.
		"""

		h = logging.StreamHandler(stream=stream)
		h.setLevel(level)
		h.setFormatter(self._formatter)
		return h

	def add_handler(self, handler):

		"""
		desc:
			Adds a handler.
		"""

		self._logger.addHandler(handler)

	def remove_handler(self, handler):

		"""
		desc:
			Removes a handler.
		"""

		self._logger.removeHandler(handler)

	@property
	def debug_mode(self):

		return '--debug' in sys.argv or '-d' in sys.argv

	@property
	def debug(self):

		return self._logger.debug

	@property
	def info(self):

		return self._logger.info

	@property
	def warning(self):

		return self._logger.warning

	@property
	def error(self):

		return self._logger.error

	@property
	def critical(self):

		return self._logger.critical

	@property
	def name(self):

		return self._logger.name

	@name.setter
	def name(self, name):

		self._logger.name = name

	@property
	def started(self):

		return hasattr(self, u'_logger')


# A singleton instance
oslogger = OSLogger()



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp.backend import Backend, configurable
from libopensesame.exceptions import osexception


class Sampler(Backend):

	"""
	desc: |
		The `Sampler` class provides functionality to play sound samples. You
		generally create a `Sampler` object with the `Sampler()` factory
		function, as described in the section
		[Creating a Sampler](#creating-a-sampler).

		__Example:__

		~~~ .python
		src = pool['bark.ogg']
		my_sampler = Sampler(src, volume=.5)
		my_sampler.play()
		~~~

		[TOC]

		## Things to know

		### Creating a Sampler

		You generally create a `Sampler` with the `Sampler()` factory function,
		which takes the full path to a sound file as the first argument.

		~~~ .python
		src = pool['bark.ogg']
		my_sampler = Sampler(src)
		~~~

		Optionally, you can pass [Playback keywords](#playback-keywords) to
		`Sampler()` to set the default behavior:

		~~~ .python
		src = pool['bark.ogg']
		my_sampler = Sampler(src, volume=.5)
		~~~

		### Sampling rate

		If you find that your sample plays to slowly (low pitch) or too quickly
		(high pitch), make sure that the sampling rate of your sample matches
		the sampling rate of the sampler back-end as specified under back-end
		settings.

		### Supported file formats

		Sound files in `.wav` and `.ogg` format are supported. If you need to
		convert samples from a different format, you can use
		[Audacity](http://sourceforge.net/projects/audacity/).

		### Backwards incompatible changes from 2.9 to 3.0

		The following are now properties, as described under
		[playback keywords]:

		- [sampler.block]
		- [sampler.duration]
		- [sampler.fade_in]
		- [sampler.pan]
		- [sampler.pitch]
		- [sampler.volume]

		Therefore, the following will no longer work:

		~~~ .python
		sampler.volume(.5)
		~~~

		And has to be changed to:

		~~~ .python
		sampler.volume = .5
		~~~

		### Playback keywords

		Functions that accept `**playback_args` take the following keyword
		arguments:

		- `volume` specifies a volume between `0.0` (silent) and `1.0`
		  (maximum).
		- `pitch` specifies a pitch (or playback speed), where values > 1
		  indicate a higher pitch, and values < 1 indicate a lower pitch.
		- `pan` specifies a panning, where values < 0 indicate panning to the
		  left, and values > 0 indicate panning to the right. Alternatively, you
		  can set pan to 'left' or 'right' to play only a single channel.
		- `duration` specifies the duration of the the sound in milliseconds, or
		  is set to `0` or `None` to play the full sound.
		- `fade_in` specifies the fade-in time (or attack) of the sound, or is
		  set to `0` or `None` to disable fade-in.
		- `block` indicates whether the experiment should block (`True`) during
		  playback or not (`False`).

		~~~ .python
		src = pool['bark.ogg']
		my_sampler = Sampler(src)
		my_sampler.play(volume=.5, pan='left')
		~~~

		Playback keywords only affect the current operation (except when passed
		to [sampler.\_\_init\_\_][__init__]). To change the behavior for all
		subsequent operations, set the playback properties directly:

		~~~ .python
		src = pool['bark.ogg']
		my_sampler = Sampler(src)
		my_sampler.volume = .5
		my_sampler.pan = 'left'
		my_sampler.play()
		my_sampler.play()
		~~~

		Or pass the playback keywords to [sampler.\_\_init\_\_][__init__]:

		~~~ .python
		src = pool['bark.ogg']
		my_sampler = Sampler(src, volume=.5, pan='left')
		my_sampler.play()
		my_sampler.play()
		~~~
	"""

	def __init__(self, experiment, src, **playback_args):

		"""
		visible: False

		desc: |
			Constructor to create a new SAMPLER object. You do not generally
			call this constructor directly, but use the `Sampler()` function,
			which is described here: [/python/sampler/]().

		arguments:
			experiment:
				desc:	The experiment object.
				type:	experiment
			src:
				desc:	The full path to a `.wav` or `.ogg` file.
				type:	[unicode, str]

		keyword-dict:
			playback_args:
				Optional [playback keywords] that will be used as the default
				for this SAMPLER object.

		example: |
			src = pool[u'my_sound.ogg']
			my_sampler = Sampler(src, volume=.5)
		"""

		self.experiment = experiment
		Backend.__init__(self, configurables={
			u'volume': self.assert_numeric,
			u'pan': self.assert_pan,
			u'pitch': self.assert_numeric,
			u'duration': self.assert_numeric_or_None,
			u'fade_in': self.assert_numeric_or_None,
			u'block': self.assert_bool,
		}, **playback_args)

	def default_config(self):

		return {
			u'volume' 			: 1.,
			u'pan'				: 0.,
			u'pitch'			: 1.,
			u'duration'			: None,
			u'fade_in'			: None,
			u'block'			: False,
		}

	def assert_pan(self, key, val):

		if not isinstance(val, int) and not isinstance(val, float) \
			and val not in [u'left', u'right']:
			raise osexception(
				u'pan should be numeric, \'left\', or \'right\', not %s' % val)

	@configurable
	def play(self, **playback_args):

		"""
		desc:
			Plays the sound.

		keyword-dict:
			playback_args:
				Optional [playback keywords] that will be used for this call to
				[sampler.play]. This does not affect subsequent operations.

		example: |
			src = pool[u'my_sound.ogg']
			my_sampler = Sampler(src)
			my_sampler.play(pitch=.5, block=True)
		"""

		raise NotImplementedError()

	def stop(self):

		"""
		desc:
			Stops the currently playing sound (if any).

		example: |
			src = pool[u'my_sound.ogg']
			my_sampler = Sampler(src)
			my_sampler.play()
			sleep(100)
			my_sampler.stop()
		"""

		raise NotImplementedError()

	def pause(self):

		"""
		desc:
			Pauses playback (if any).

		example: |
			src = pool[u'my_sound.ogg']
			my_sampler = Sampler(src)
			my_sampler.play()
			sleep(100)
			my_sampler.pause()
			sleep(100)
			my_sampler.resume()
		"""

		raise NotImplementedError()

	def resume(self):

		"""
		desc:
			Resumes playback (if any).

		example: |
			src = pool[u'my_sound.ogg']
			my_sampler = Sampler(src)
			my_sampler.play()
			sleep(100)
			my_sampler.pause()
			sleep(100)
			my_sampler.resume()
		"""

		raise NotImplementedError()

	def is_playing(self):

		"""
		desc:
			Checks if a sound is currently playing.

		returns:
			desc:	True if a sound is playing, False if not.
			type:	bool

		example: |
			src = pool[u'my_sound.ogg']
			my_sampler = Sampler(src)
			my_sampler.play()
			sleep(100)
			if my_sampler.is_playing():
				print('The sampler is still playing!')
		"""

		raise NotImplementedError()

	def wait(self):

		"""
		desc:
			Blocks until the sound has finished playing or returns right away if
			no sound is playing.

		example: |
			src = pool[u'my_sound.ogg']
			my_sampler = Sampler(src)
			my_sampler.play()
			my_sampler.wait()
			print('The sampler is finished!')
		"""

		raise NotImplementedError()

	# Deprecated functions

	def stop_after(self, ms):

		"""
		visible:	False
		desc:		deprecated
		"""

		warnings.warn(
			u'sampler.stop_after() has been deprecated. '
			'Use sampler.duration instead.',
			DeprecationWarning
		)
		self.duration = ms

	@staticmethod
	def init_sound(experiment):

		"""
		visible: False

		desc:
			Initializes the pygame mixer before the experiment begins.

		arguments:
			experiment:
				desc:	The experiment object.
				type:	experiment
		"""

		raise NotImplementedError()

	@staticmethod
	def close_sound(experiment):

		"""
		visible: False

		desc:
			Closes the mixer after the experiment is finished.

		arguments:
			experiment:
				desc:	The experiment object.
				type:	experiment
		"""

		raise NotImplementedError()


# Non PEP-8 alias for backwards compatibility
sampler = Sampler



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from libopensesame import item, widgets
from libqtopensesame.items.qtautoplugin import qtautoplugin


class form_multiple_choice(item.item):

	description = u'A simple multiple choice item'
	lazy_init = False

	def reset(self):

		"""
		desc:
			Initialize plug-in.
		"""

		self.var.options = u'Yes\nNo\nMaybe'
		self.var.question = u'Your question'
		self.var.form_title = u'Form title'
		self.var.form_var = u'response'
		self.var.advance_immediately = u'yes'
		self.var.allow_multiple = u'yes'
		self.var.button_text = u'Ok'
		self.var.timeout = u'infinite'
		self.var.spacing = 10
		self.var.margins = u'50;50;50;50'
		self.var._theme = u'gray'

	def run(self):

		"""Run the item"""

		# Parse the option list
		option_list = self.var.options.split(u'\n')  # split by return
		# Filter out empty options
		option_list = list(filter(lambda option: option != u'', option_list))
		if not option_list:
			raise osexception(
				(
					u'You must specify at least one response option in'
					u'form_multiple_choice item "%s"'
				)
				% self.name
			)
		# Determine whether a button is shown and determine the number of rows
		rows = len(option_list) + 2
		if (
			self.var.advance_immediately == u'no' or
			self.var.allow_multiple == u'yes'
		):
			show_button = True
			click_accepts = False
			rows += 1
		else:
			show_button = False
			click_accepts = True
		# Determine the group for the checkboxes
		if self.var.allow_multiple == u'no':
			group = u'response_group'
		else:
			group = None
		# The variable in which the response is stored
		var = self.var.form_var
		# Build the form
		try:
			margins = [
				float(i) for i in safe_decode(self.var.margins).split(u';')
			]
		except (ValueError, TypeError):
			raise osexception(
				_(u'margins should be numeric values separated by a semi-colon')
			)
		if self.var.timeout == u'infinite':
			timeout = None
		else:
			timeout = self.var.timeout
		form = widgets.form(
			self.experiment,
			cols=1,
			rows=rows,
			spacing=self.var.spacing,
			margins=margins,
			theme=self.var._theme,
			item=self,
			timeout=timeout,
			clicks=self.var.form_clicks == u'yes'
		)
		form.set_widget(widgets.label(form, self.var.form_title), (0, 0))
		form.set_widget(widgets.label(form, self.var.question), (0, 1))
		i = 2
		for option in option_list:
			form.set_widget(
				widgets.checkbox(
					form,
					option,
					group=group,
					click_accepts=click_accepts,
					var=var
				),
				(0, i)
			)
			i += 1
		if show_button:
			form.set_widget(widgets.button(form, self.var.button_text), (0, i))
		# Go!
		form._exec()

	def var_info(self):

		"""
		Return a list of dictionaries with variable descriptions

		Returns:
		A list of (name, description) tuples
		"""

		return (
			item.item.var_info(self) +
			[(self.var.form_var, u'[Depends on response]')]
		)


class qtform_multiple_choice(form_multiple_choice, qtautoplugin):

	"""GUI controls"""

	def __init__(self, name, experiment, string=None):

		"""
		Constructor

		Arguments:
		name		--	The name of the item.
		experiment	--	The experiment instance.

		Keyword arguments:
		string		--	A definition string. (default=None)
		"""

		form_multiple_choice.__init__(self, name, experiment, string)
		qtautoplugin.__init__(self, __file__)
		self.custom_interactions()

	def apply_edit_changes(self):

		"""Apply the controls"""

		if not qtautoplugin.apply_edit_changes(self) or self.lock:
			return False
		self.custom_interactions()
		return True

	def custom_interactions(self):

		"""
		The advance_immediately option is not applicable if multiple items can
		be selected.
		"""

		self.checkbox_advance_immediately.setEnabled(
			self.var.get(u'allow_multiple') == u'no'
		)
		self.line_edit_button_text.setEnabled(
			self.var.get(u'allow_multiple') == u'yes' or
			self.var.get(u'advance_immediately') == u'no'
		)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp.backend import Backend, configurable
from libopensesame.exceptions import osexception
import warnings


class Mouse(Backend):

	"""
	desc: |
		The `Mouse` class is used to collect mouse input. You generally create a
		`Mouse` object with the `Mouse()` factory function, as described in the
		section [Creating a Mouse](#creating-a-mouse).

		__Example:__

		~~~ .python
		# Draw a 'fixation-dot mouse cursor' until a button is clicked
		my_mouse = Mouse()
		my_canvas = Canvas()
		while True:
			button, position, timestamp = my_mouse.get_click(timeout=20)
			if button is not None:
				break
			(x,y), time = my_mouse.get_pos()
			my_canvas.clear()
			my_canvas.fixdot(x, y)
			my_canvas.show()
		~~~

		[TOC]

		## Things to know

		### Creating a Mouse

		You generally create a `Mouse` with the `Mouse()` factory function:

		~~~ .python
		my_mouse = Mouse()
		~~~

		Optionally, you can pass [Response keywords](#response-keywords) to
		`Mouse()` to set the default behavior:

		~~~ .python
		my_mouse = Mouse(timeout=2000)
		~~~

		### Coordinates

		- When *Uniform coordinates* is set to 'yes', coordinates are
		  relative to the center of the display. That is, (0,0) is the center.
		  This is the default as of OpenSesame 3.0.0.
		- When *Uniform coordinates* is set to 'no', coordinates are relative to
		  the top-left of the display. That is, (0,0) is the top-left. This was
		  the default in OpenSesame 2.9.X and earlier.

		### Button numbers

		Mouse buttons are numbered as follows:

		1. Left button
		2. Middle button
		3. Right button
		4. Scroll up
		5. Scroll down

		### Touch screens

		When working with a touch screen, a touch is registered as button 1
		(left button).

		### Response keywords

		Functions that accept `**resp_args` take the following keyword
		arguments:

		- `timeout` specifies a timeout value in milliseconds, or is set to
		  `None` to disable the timeout.
		- `buttonlist` specifies a list of buttons that are accepted, or is set
		  to `None` accept all buttons.
		- `visible` indicates whether the mouse cursor becomes visible when a
		  click is collected (`True` or `False`). To immediately change cursor
		  visibility, use [mouse.show_cursor].

		~~~ .python
		# Get a left or right button press with a timeout of 3000 ms
		my_mouse = Mouse()
		button, time = my_mouse.get_click(buttonlist=[1,3], timeout=3000)
		~~~

		Response keywords only affect the current operation (except when passed
		to [mouse.\_\_init\_\_][__init__]). To change the behavior for all
		subsequent operations, set the response properties directly:

		~~~ .python
		# Get two left or right presses with a 5000 ms timeout
		my_mouse = Mouse()
		my_mouse.buttonlist = [1,3]
		my_mouse.timeout = 5000
		button1, time1 = my_mouse.get_click()
		button2, time2 = my_mouse.get_click()
		~~~

		Or pass the response keywords to [mouse.\_\_init\_\_][__init__]:

		~~~ .python
		# Get two left or right presses with a 5000 ms timeout
		my_mouse = Mouse(buttonlist=[1,3], timeout=5000)
		button1, time1 = my_mouse.get_click()
		button2, time2 = my_mouse.get_click()
		~~~
	"""

	def __init__(self, experiment, **resp_args):

		"""
		visible: False

		desc: |
			Constructor to create a new `Mouse` object. You do not generally
			call this constructor directly, but use the `Mouse()` function,
			which is described here: [/python/common/]().

		arguments:
			experiment:
				desc:		The experiment object.
				type:		experiment

		keyword-dict:
			resp_args:
				Optional [response keywords] that will be used as the default
				for this `Mouse` object.

		example: |
			my_mouse = Mouse(buttonlist=[1, 2], timeout=2000)
		"""

		self.experiment = experiment
		self._cursor_shown = False
		Backend.__init__(self, configurables={
			u'timeout' : self.assert_numeric_or_None,
			u'buttonlist' : self.assert_list_or_None,
			u'visible' : self.assert_bool,
			}, **resp_args
		)

	def set_config(self, **cfg):

		# Add synonyms to buttonlist
		if u'buttonlist' in cfg and isinstance(cfg[u'buttonlist'], list):
			try:
				cfg[u'buttonlist'] = \
					[int(button) for button in cfg[u'buttonlist']]
			except:
				raise osexception(
					u"buttonlist must be a list of numeric values, or None")
		Backend.set_config(self, **cfg)

	def default_config(self):

		return {
			u'timeout' 			: None,
			u'buttonlist'		: None,
			u'visible'			: False
			}

	def show_cursor(self, show=True):

		"""
		desc: |
			Immediately changes the visibility of the mouse cursor.

			__Note:__ In most cases, you will want to use the `visible`
			keyword, which changes the visibility during response collection,
			that is, while `mouse.get_click()` is called. Calling 
			`show_cursor()` will not implicitly change the value of `visible`, 
			which can lead to the somewhat unintuitive behavior that the cursor
			is hidden as soon as `get_click()` is called.
			

		keywords:
			show:
				desc: 	Indicates whether the cursor is shown (True) or hidden
						(False).
				type:	bool
		"""

		self._cursor_shown = show

	def set_pos(self, pos=(0,0)):

		"""
		desc: |
			Sets the position of the mouse cursor.

			__Warning:__ `set_pos()` is unreliable and will silently fail on
			some systems.

		keywords:
			pos:
				desc:	An (x,y) tuple for the new mouse coordinates.
				type:	tuple

		example: |
			my_mouse = Mouse()
			my_mouse.set_pos(pos=(0,0))
		"""

		pass

	@configurable
	def get_click(self, **resp_args):

		"""
		desc:
			Collects a mouse click.

		keyword-dict:
			resp_args:
				Optional [response keywords] that will be used for this call to
				[mouse.get_click]. This does not affect subsequent operations.

		returns:
			desc:			A (button, position, timestamp) tuple. The button
							and position are `None` if a timeout occurs.
							Position is an (x, y) tuple in screen coordinates.
			type:			tuple

		example: |
			my_mouse = Mouse()
			button, (x, y), timestamp = my_mouse.get_click(timeout=5000)
			if button is None:
				print('A timeout occurred!')
		"""

		raise NotImplementedError()

	@configurable
	def get_click_release(self, **resp_args):

		"""
		desc: |
			*New in v3.2.0*

			Collects a mouse-click release.

			*Important:* This function is currently not implemented for the
			*psycho* backend.

		keyword-dict:
			resp_args:
				Optional [response keywords] that will be used for this call to
				[mouse.get_click_release]. This does not affect subsequent
				operations.

		returns:
			desc:			A (button, position, timestamp) tuple. The button
							and position are `None` if a timeout occurs.
							Position is an (x, y) tuple in screen coordinates.
			type:			tuple

		example: |
			my_mouse = Mouse()
			button, (x, y), timestamp = my_mouse.get_click_release(timeout=5000)
			if button is None:
				print('A timeout occurred!')
		"""

		raise NotImplementedError()

	def get_pos(self):

		"""
		desc:
			Returns the current position of the cursor.

		returns:
			desc:	A (position, timestamp) tuple.
			type:	tuple

		example: |
			my_mouse = Mouse()
			(x, y), timestamp = my_mouse.get_pos()
			print('The cursor was at (%d, %d)' % (x, y))
		"""

		raise NotImplementedError()

	def get_pressed(self):

		"""
		desc:
			Returns the current state of the mouse buttons. A True value means
			the button is currently being pressed.

		returns:
			desc:	A (button1, button2, button3) tuple of boolean values.
			type:	tuple.

		example: |
			my_mouse = Mouse()
			buttons = my_mouse.get_pressed()
			b1, b2, b3 = buttons
			print('Currently pressed mouse buttons: (%d,%d,%d)' % (b1,b2,b3))
		"""

		raise NotImplementedError()

	def flush(self):

		"""
		desc:
			Clears all pending input, not limited to the mouse.

		returns:
			desc:	True if a button had been clicked (i.e., if there was
					something to flush) and False otherwise.
			type:	bool

		example: |
			my_mouse = Mouse()
			my_mouse.flush()
			button, position, timestamp = my_mouse.get_click()
		"""

		raise NotImplementedError()

	def synonyms(self, button):

		"""
		visible: False

		desc:
			Gives a list of synonyms for a mouse button. For example, 1 and
			'left_button' are synonyms.

		arguments:
			button:
				desc:	A button value.
				type:	[int, str, unicode]

		returns:
			desc:	A list of synonyms.
			type:	list
		"""

		button_map = [
			(1, u"left_button"),
			(2, u"middle_button"),
			(3, u"right_button"),
			(4, u"scroll_up"),
			(5, u"scroll_down")
			]
		for bm in button_map:
			if button in bm:
				return bm
		return []

	# Deprecated functions

	def set_buttonlist(self, buttonlist=None):

		"""
		visible:	False
		desc:		deprecated
		"""

		warnings.warn(u'mouse.set_buttonlist() has been deprecated. '
			'Use mouse.buttonlist instead.', DeprecationWarning)
		self.buttonlist = buttonlist

	def set_timeout(self, timeout=None):

		"""
		visible:	False
		desc:		deprecated
		"""

		warnings.warn(u'mouse.set_timeout() has been deprecated. '
			'Use mouse.timeout instead.', DeprecationWarning)
		self.timeout = timeout

	def set_visible(self, visible=True):

		"""
		visible:	False
		desc:		deprecated
		"""

		warnings.warn(u'mouse.set_visible() has been deprecated. '
			'Use mouse.visible instead.', DeprecationWarning)
		self.visible = visible


# Non PEP-8 alias for backwards compatibility
mouse = Mouse



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import operator


class cistr(str):
	
	"""
	desc:
		A class for case-insensitive string comparisons. This is used to speed
		up the item store, which gets a lot of look ups for item names. This is
		not a full emulation of the str protocol, only those parts that are used
		by the item store.
	"""

	def __init__(self, s):
		
		if py3:			
			str.__init__(s)
		else:
			str.__init__(self, s)
		self._lower = s.lower()
		self._hash = hash(self._lower)

	def __eq__(self, other):
		
		try:
			return other._hash == self._hash
		except AttributeError:
			if hasattr(other, u'lower'):
				return other.lower() == self._lower
		return other == self._lower
		
	def __ne__(self, other):
		
		try:
			return other._hash != self._hash
		except AttributeError:
			if hasattr(other, u'lower'):
				return other.lower() != self._lower
		return other != self._lower				
		
	def __hash__(self):
		
		return self._hash



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.widgets.tree_base_item import tree_base_item
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'tree_recursion_error_item', category=u'core')


class tree_recursion_error_item(tree_base_item):

	"""
	desc:
		A placeholder item that is shown when there is a recursion in the
		experiment that prevents the item tree from being built.
	"""

	def __init__(self, main_window):

		"""
		desc:
			Constructor.

		arguments:
			main_window:
				desc:	The main-window object.
				type:	qtopensesame
		"""

		super(tree_recursion_error_item, self).__init__()
		self.setup(main_window)
		self.setText(0, _(u'Recursion detected'))
		self.setIcon(0, self.theme.qicon(u'dialog-error'))
		self._droppable = False
		self._draggable = False
		self.name = u'__recursion_error__'

	def open_tab(self):

		self.main_window.tabwidget.open_general_script()



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.

# About

This module is used to maintain configuration settings. When this module is
first loaded, a single instance of the `config` class is instantiated, whicn is
subsequently used for all configuration getting and setting (i.e. a singleton
design pattern).

	from libqtopensesame.misc.config import cfg
	cfg.my_setting = 'my_value' # set
	print(cfg.my_setting) # get
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from qtpy import QtCore
import libopensesame.misc
from libopensesame.oslogging import oslogger
import platform
import sys


DEFAULT_CONFIG = {
	u"cfg_ver" : 0,
	u"_initial_window_geometry" : QtCore.QByteArray(),
	u"_initial_window_state" : QtCore.QByteArray(),
	u"auto_update_check" : True,
	u"auto_response" : False,
	u"default_logfile_folder" : libopensesame.misc.home_folder(),
	u"default_pool_folder" : libopensesame.misc.home_folder(),
	u"disabled_plugins" : "",
	u"disabled_extensions" : "",
	u"file_dialog_path" : "",
	u"file_pool_size_warning" : 104857600,
	u"loop_wizard" : None,
	u"onetabmode" : False,
	u"quick_run_logfile": u"quickrun.csv",
	u"recent_files" : u"",
	u"reset_console_on_experiment_start" : True,
	u"shortcut_itemtree" : u"Ctrl+1",
	u"shortcut_tabwidget" : u"Ctrl+2",
	u"shortcut_stdout" : u"Ctrl+3",
	u"shortcut_pool" : u"Ctrl+4",
	u"shortcut_copy_clipboard_unlinked" : u"Ctrl+C",
	u"shortcut_copy_clipboard_linked" : u"Ctrl+Shift+C",
	u"shortcut_paste_clipboard" : u"Ctrl+V",
	u"shortcut_delete" : u"Del",
	u"shortcut_permanently_delete" : u"Shift+Del",
	u"shortcut_context_menu" : u"+",
	u"shortcut_rename" : u"F2",
	u"shortcut_edit_runif" : u"F3",
	u"style" : u"",
	u"theme" : u"default",
	u"toolbar_size" : 32,
	u"toolbar_text" : False,
	u"runner" : u"multiprocess",
	u"opensesamerun_exec" : u"",
	u"start_drag_delay" : 300,
	u"pos" : QtCore.QPoint(200, 200),
	u"size" : QtCore.QSize(1000, 600),
	u"url_website" : u"http://www.cogsci.nl/opensesame",
	u"url_facebook" : u"http://www.facebook.com/cognitivescience",
	u"url_twitter" : u"http://www.twitter.com/cogscinl",
	u'sketchpad_placeholder_color' : u'#00FF00',
	u'sketchpad_grid_color' : u'#00FF00',
	u'sketchpad_grid_thickness_thin' : 2,
	u'sketchpad_grid_thickness_thick' : 4,
	u'sketchpad_grid_opacity' : 32,
	u'sketchpad_preview_color' : u'#00FF00',
	u'sketchpad_preview_penwidth' : 2,
	u'mode': u'default',
	u'locale' : u'',
	}
DEFAULT_CONFIG_LINUX = {
	u"theme" : u"gnome"
	}
DEFAULT_CONFIG_MAC = {}
DEFAULT_CONFIG_WINDOWS = {}


class config(object):

	def __init__(self):

		"""
		desc:
			Constructor.
		"""

		self.reset()

	def __str__(self):

		return self.__unicode__() if py3 else safe_encode(self.__unicode__())

	def __unicode__(self):

		s = u''
		for key, val in self.config.items():
			if not isinstance(val, basestring) and not isinstance(val, int) \
				and not isinstance(val, float):
				val = type(val)
			s += u'%s: %s\n' % (key, val)
		return s

	def __getattr__(self, setting):

		"""
		desc:
			A getter for settings, to allow for easy access

		arguments:
			setting:	The setting to get
		"""

		if setting not in self.config:
			raise osexception(u'The setting "%s" does not exist' \
				% setting)
		return self.config[setting]

	def __setattr__(self, setting, value):

		"""
		desc:
			A setter for settings, to allow for easy access

		argumentsL
			setting: 	The setting to set
			value:		The value to set
		"""

		if setting not in self.config:
			raise osexception(u'The setting "%s" does not exist' \
				% setting)
		self.config[setting] = value
		self.config[u'cfg_ver'] += 1

	def __setitem__(self, setting, value):

		"""
		desc:
			Emulate dict API.
		"""

		self.__setattr__(setting, value)

	def __getitem__(self, setting):

		"""
		desc:
			Emulate dict API.
		"""

		return self.__getattr__(setting)

	def __contains__(self, setting):

		"""
		returns:
			True if setting exists, False otherwise.
		"""

		return setting in self.config

	def register(self, setting, default):

		"""
		desc:
			Registers a new setting, if it doesn't already exist.

		arguments:
			setting:	The setting name.
			default:	A default value, which is used if the setting doesn't
						already exist.
		"""

		qsettings = QtCore.QSettings(u"cogscinl", self.mode)
		qsettings.beginGroup(u"MainWindow")
		if '--start-clean' in sys.argv:
			self.config[setting] = default
		elif setting not in self.config:
			value = qsettings.value(setting, default)
			value = self.type_qvariant(value, default)
			self.config[setting] = value
		qsettings.endGroup()

	def parse_cmdline_args(self, args):

		"""
		desc:
			Apply settings that were specified on the command line. The expected
			format is as follows: [name]=[val];[name]=[val];...

		arguments:
			args:	The string of command line arguments
		"""

		if args is None:
			return
		for arg in args.split(u";"):
			a = arg.split(u"=")
			if len(a) == 2:
				# Automagically determine the data type
				if a[1] == u"True":
					val = True
				elif a[1] == u"False":
					val = False
				else:
					try:
						val = int(a[1])
					except:
						try:
							val = float(a[1])
						except:
							val = a[1]
				# Apply the argument
				try:
					self.__setattr__(a[0], val)
					oslogger.debug(u"%s = %s" % (a[0], val))
				except:
					oslogger.error(u"Failed to parse argument: %s" % arg)

	def type_qvariant(self, value, default):

		"""
		desc:
			Typecasts a value to a normal type that matches the type of a
			default value. This is necessary, because under some combinations
			of Python 2/3 and PyQt 4/5 settings are returned as QVariant
			objects, whereas on other combinations the type casting occurs
			automatically.

		arguments:
			value:		A value.
			default:	A default value.

		returns:
			A value.
		"""

		if hasattr(QtCore, u'QPyNullVariant') and \
				isinstance(value, QtCore.QPyNullVariant):
			return default
		if isinstance(default, bool):
			if isinstance(value, basestring):
				return value == u'true'
			try:
				return bool(value)
			except:
				oslogger.error('Failed to convert %s' % value)
				return default
		if isinstance(default, int):
			try:
				return int(value)
			except:
				oslogger.error('Failed to convert %s' % value)
				return default
		if isinstance(default, float):
			try:
				return float(value)
			except:
				oslogger.error('Failed to convert %s' % value)
				return default
		return value

	def restore(self, mode):

		"""
		desc:
			Restore settings from a QSettings.
		"""

		oslogger.debug('restoring config profile {}'.format(mode))
		qsettings = QtCore.QSettings(u"cogscinl", mode)
		qsettings.beginGroup(u'MainWindow')
		for setting, default in self.config.items():
			try:
				value = qsettings.value(setting, default)
			except TypeError:
				continue
			value = self.type_qvariant(value, default)
			self.config[setting] = value
		self.mode = mode
		qsettings.endGroup()

	def save(self):

		"""
		desc:
			Save settings to a QSettings.
		"""

		oslogger.debug('saving config profile {}'.format(self.mode))
		qsettings = QtCore.QSettings(u"cogscinl", self.mode)
		qsettings.beginGroup(u'MainWindow')
		for setting, value in self.config.items():
			if setting != u"cfg_ver":
				qsettings.setValue(setting, value)
		qsettings.endGroup()

	def clear(self):

		"""
		desc:
			Clears all settings.
		"""

		qsettings = QtCore.QSettings(u"cogscinl", self.mode)
		qsettings.beginGroup(u'MainWindow')
		qsettings.clear()
		qsettings.endGroup()

	def version(self):

		"""
		desc:
			Gets the current version of the config.

		Returns:
			The config version.
		"""

		return self.cfg_ver

	def reset(self):

		"""
		desc:
			Resets the configution back to default.
		"""

		object.__setattr__(self, u'config', {})
		self.config.update(DEFAULT_CONFIG)
		if platform.system() == u"Windows":
			self.config.update(DEFAULT_CONFIG_WINDOWS)
		elif platform.system() == u"Darwin":
			self.config.update(DEFAULT_CONFIG_MAC)
		elif platform.system() == u"Linux":
			self.config.update(DEFAULT_CONFIG_LINUX)

	def nuke(self):

		"""
		desc:
			Clears the config.
		"""

		self.reset()
		self.clear()
		self.save()


# Create a singleton config instance
cfg = config()



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *

from qtpy import QtCore, QtWidgets
from libqtopensesame.dialogs.base_dialog import base_dialog
from libqtopensesame.misc.config import cfg

class update(base_dialog):

	"""
	desc:
		A simple text-input dialog.
	"""

	def __init__(self, main_window, msg=None):

		"""
		desc:
			Constructor.

		arguments:
			main_window:	The main window object.
			msg:			A text message.
		"""

		super(update, self).__init__(main_window, ui=u'dialogs.update_dialog')
		self.ui.textedit_notification.setHtml(msg)
		self.adjustSize()

	def exec_(self):

		"""
		desc:
			Executes the dialog.
		"""

		self.ui.checkbox_auto_check_update.setChecked(cfg.auto_update_check)
		super(update, self).exec_()
		cfg.auto_update_check = self.ui.checkbox_auto_check_update.isChecked()
		self.main_window.update_preferences_tab()



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.base_subcomponent import base_subcomponent


class qtitem_splitter(base_subcomponent, QtWidgets.QSplitter):

	"""
	desc:
		Implements a splitter for the edit and script view in an item tab. This
		custom is mostly necessary, because the default QSplitter.sizeHint()
		is too large.
	"""

	def __init__(self, item):

		"""
		desc:
			Constructor.

		item:
			desc:	The item.
			type:	qtitem
		"""

		super(qtitem_splitter, self).__init__(
			QtCore.Qt.Vertical,
			item.main_window
		)
		self.item = item
		self.setup(item.main_window)
		self.addWidget(self.item._edit_widget)
		self.addWidget(self.item._script_frame)
		self.splitterMoved.connect(self._on_splitter_moved)

	def _on_splitter_moved(self, pos, index):

		self.splitterMoved.disconnect()
		self.item.init_script_widget()

	def minimumSizeHint(self):

		"""
		returns:
			type:	QSize
		"""

		return QtCore.QSize(100, 100)

	def sizeHint(self):

		"""
		returns:
			type:	QSize
		"""

		return QtCore.QSize(100, 100)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame import type_check
from libopensesame.exceptions import osexception
from openexp.canvas import canvas
from openexp.mouse import mouse
from openexp.keyboard import keyboard
from openexp._keyboard.keybabel import KeyBabel
from libopensesame.widgets.widget_factory import WidgetFactory


class Form(object):

	"""
	desc: |
		The `Form` is a container for widgets, such as labels, etc. If you use
		the FORM_BASE plug-in in combination with OpenSesame script, you do not
		need to explicitly create a `Form` object. However, if you use Python
		inline code, you do.

		__Example__:

		~~~ .python
		form = Form()
		label = Label(text='label)
		form.set_widget(label, (0,0))
		form._exec()
		~~~

		[TOC]
	"""

	def __init__(self, experiment, cols=2, rows=2, spacing=10,
		margins=(100, 100, 100, 100), theme=u'gray', item=None, timeout=None,
		clicks=False, validator=None):

		"""
		desc: |
			Constructor to create a new `Form` object. You do not generally
			call this constructor directly, but use the `Form()` factory
			function, which is described here: [/python/common/]().


		arguments:
			experiment:
				desc:	An OpenSesame experiment.
				type:	experiment

		keywords:
			cols:
				desc:	The number of columns (as int) or a list that specifies
						the number and relative size of the columns. For
						example, `[1,2,1]` will create 3 columns where the
						middle one is twice as large as the outer ones.
				type:	[int, list]
			rows:
				desc:	Analogous to `cols`, but for the rows.
				type:	[int, list]
			spacing:
				desc:	The amount of empty space between widgets (in pixels).
				type:	int
			margins:
				desc:	The amount of empty space around the form. This is
						specified as a list, like so [top-margin, right-margin,
						bottom-margin, left-margin].
				type:	list
			theme:
				desc:	The theme for the widgets.
				type:	[str, unicode]
			item:
				desc:	The item of which the form is part.
				type:	[item, NoneType]
			timeout:
				desc:	A timeout value in milliseconds, or None if no timeout
						exists.
				type:	[int, float, NoneType]
			clicks:
				desc:	If enabled, an auditory click is played on user
						interactions. This can help to make interactions feel
						smoother if there is some visual lag.
				type:	bool
			validator:
				desc:	A function that takes no arguments and returns True if
						the form is successfully validated, and False if not.
				type: 	[FunctionType, NoneType]
		"""

		# Normalize the column and row sizes so that they add up to 1
		if isinstance(cols, int):
			self.cols = [1./cols]*cols
		else:
			cols = type_check.float_list(cols, u'form columns', min_len=1)
			self.cols = [float(c)/sum(cols) for c in cols]
		if isinstance(rows, int):
			self.rows = [1./rows]*rows
		else:
			rows = type_check.float_list(rows, u'form rows', min_len=1)
			self.rows = [float(r)/sum(rows) for r in rows]
		self.experiment = experiment
		self.item = item if item is not None else experiment
		self.timeout = timeout
		self.width = experiment.var.width
		self.height = experiment.var.height
		self.spacing = spacing
		self.clicks = clicks
		self.margins = type_check.float_list(margins, u'form margins',
			min_len=4, max_len=4)
		n_cells = len(self.cols)*len(self.rows)
		self.widgets = [None]*n_cells
		self.span = [(1, 1)]*n_cells
		self._validator = (lambda: True) if validator is None else validator
		self.canvas = canvas(
			self.experiment,
			color=self.item.var.foreground,
			background_color=self.item.var.background
		)
		# Dynamically load the theme object
		theme_mod = __import__(
			u'libopensesame.widgets.themes.%s' % theme, fromlist=[u'dummy'])
		theme_cls = getattr(theme_mod, theme)
		self.theme_engine = theme_cls(self)

	def __len__(self):

		"""
		visible: False

		desc:
			Implements the `len()` syntax.

		returns:
			desc:	The number of widgets in the form.
			type:	int
		"""

		return sum([w is not None for w in self.widgets])

	def render(self):

		"""
		desc:
			Shows the form canvas without any user interaction.
		"""

		self.canvas.show()

	def _exec(self, focus_widget=None):

		"""
		desc:
			Executes the form.

		keywords:
			focus_widget:
				desc:	A widget that is in the form and should receive a
						virtual mouse click when the form is opened. This allows
						you to activate a text_input right away, for example, so
						that the user doesn't have to click on it anymore.
				type:	[widget, NoneType]

		returns:
			desc:	Gives the return value of the form, which depends on how the
					user interacted with the widgets. For example, if the user
					pressed a button, the button text will be returned. If a
					timeout occurred, None will be returned.
		"""

		if isinstance(focus_widget, WidgetFactory):
			focus_widget = focus_widget.construct(self)
		if focus_widget is not None:
			focus_widget.focus = True
		if len(self) == 0:
			raise osexception(u'The form contains no widgets')
		ms = mouse(self.experiment, timeout=0)
		ms.show_cursor()
		kb = keyboard(self.experiment, timeout=0)
		kb.show_virtual_keyboard()
		babel = KeyBabel(kb)
		coroutines = {w: w.coroutine() for w in self.widgets if w is not None}
		for coroutine in coroutines.values():
			coroutine.send(None)
		self.canvas.show()
		self.start_time = None
		mousedown = False
		while True:
			if self.timed_out():
				resp = None
				break
			msg = None
			# Handle mouse clicks, including waiting until the mouse is released
			# after a mouse click.
			if mousedown:
				while any(ms.get_pressed()):
					pass
				mousedown = False
			button, xy, timestamp = ms.get_click(visible=True)
			if button is not None:
				mousedown = True
				# Switch the focus to the newly clicked widget (if any)
				widget = self.xy_to_widget(xy)
				if widget is None:
					continue
				if focus_widget is not None:
					focus_widget.focus = False
				widget.focus = True
				focus_widget = widget
				msg = {
					u'type': u'click',
					u'pos': xy,
					u'button': button,
					u'timestamp': timestamp
				}
			# Handle key presses clicks
			elif focus_widget is not None:
				key, timestamp = kb.get_key()
				modifiers = kb.get_mods()
				if key is not None:
					msg = {
						u'type': u'key',
						u'key': babel.standard_name(
							key,
							shift=u'shift' in modifiers
						),
						u'timestamp': timestamp
					}
			# Send message (if any)
			if msg is None:
				continue
			resp = coroutines[focus_widget].send(msg)
			self.canvas.show()
			if resp is not None and self._validator():
				break
		kb.show_virtual_keyboard(False)
		ms.show_cursor(False)
		for coroutine in coroutines.values():
			try:
				coroutine.send({u'type' : u'stop'})
			except StopIteration:
				pass
		self.experiment.var.form_response = resp
		return resp

	def timed_out(self):

		"""
		visible: False

		returns:
			desc:	True if a timeout occurred, False otherwise.
			type:	bool
		"""

		if self.timeout is None:
			return False
		if self.start_time is None:
			self.start_time = self.item.clock.time()
			return False
		return self.item.clock.time()-self.start_time >= self.timeout

	def cell_index(self, pos):

		"""
		desc:
			Converts a position to a cell index. A cell index corresponds to the
			number of the cell in the form, from left-to-right, top-to-bottom.

		arguments:
			pos:
				desc:	A position in the form, which can be an index (int) or a
						(column, row) tuple.
				type:	[int, tuple]

		returns:
			desc:	A cell index.
			type:	int
		"""

		if type(pos) == int:
			return pos
		if type(pos) in (tuple, list) and len(pos) == 2:
			return pos[1]*len(self.cols)+pos[0]
		raise osexception(u'%s is an invalid position in the form' % pos)

	def validate_geometry(self):

		"""
		Checks whether the form has a valid geometry.

		Exceptions:
		osexception		--	When the geometry is invalid.
		"""

		for index1 in range(len(self.widgets)):
			if self.widgets[index1] is None:
				continue
			l = self.get_cell(index1)
			colspan, rowspan = self.span[index1]
			for col in range(l[0], l[0]+colspan):
				for row in range(l[1], l[1]+rowspan):
					index2 = self.cell_index((col, row))
					if index1 == index2:
						continue
					if len(self.widgets) <= index2:
						raise osexception(
							u'The widget at position (%d, %s) falls outside of your form'
							% (l[0], l[1]))
					if self.widgets[index2] is not None:
						raise osexception(
							u'The widget at position (%d, %d) overlaps with another widget'
							% (l[0], l[1]))

	def get_cell(self, index):

		"""
		Returns the position of a widget

		Arguments:
		index -- the index of the widget

		Returns:
		A (column, row, column_span, row_span) tuple
		"""

		index = self.cell_index(index)
		col = index % len(self.cols)
		row = index // len(self.cols)
		colspan, rowspan = self.span[index]
		return col, row, colspan, rowspan

	def get_rect(self, index):

		"""
		Returns the boundary area for a given cell

		Arguments:
		index -- a cell index

		Returns:
		A (left, top, width, height) tuple
		"""

		col = index % len(self.cols)
		row = index // len(self.cols)
		colspan, rowspan = self.span[index]
		effective_width = self.width-self.margins[1]-self.margins[3]
		effective_height = self.height-self.margins[0]-self.margins[2]
		x1 = effective_width*sum(self.cols[:col])+self.spacing
		y1 = effective_height*sum(self.rows[:row])+self.spacing
		x2 = effective_width*sum(self.cols[:col+colspan])-self.spacing
		y2 = effective_height*sum(self.rows[:row+rowspan])-self.spacing
		w = x2-x1
		h = y2-y1
		if w <= 0 or h <= 0:
			raise osexception(u'There is not enough space to show some form '
				u'widgets. Please modify the form geometry!')
		x = x1+self.margins[3]
		y = y1+self.margins[0]
		if self.experiment.var.uniform_coordinates == u'yes':
			x -= self.width/2
			y -= self.height/2
		return x, y, w, h

	def set_widget(self, widget, pos, colspan=1, rowspan=1):

		"""
		desc:
			Adds a widget to the form.

		arguments:
			widget:
				desc:	The widget to add.
				type:	widget
			pos:
				desc:	The position to add the widget, which can be an index or
						a (column, row) tuple.
				type:	[int, tuple]

		keywords:
			colspan:
				desc:	The number of columns that the widget should span.
				type:	int
			rowspan:
				desc:	The number of rows that the widget should span.
				type:	int
		"""

		index = self.cell_index(pos)
		if index >= len(self.widgets):
			raise osexception(
				u'Widget position (%s, %s) is outside of the form' % pos)
		if not isinstance(colspan, int) or colspan < 1 or colspan > len(self.cols):
			raise osexception(
				u'Column span %s is invalid (i.e. too large, too small, or not a number)'
				% colspan)
		if not isinstance(rowspan, int) or rowspan < 1 or rowspan > len(self.rows):
			raise osexception(
				u'Row span %s is invalid (i.e. too large, too small, or not a number)'
				% rowspan)
		if isinstance(widget, WidgetFactory):
			widget = widget.construct(self)
		self.widgets[index] = widget
		self.span[index] = colspan, rowspan
		widget.set_rect(self.get_rect(index))

	def xy_to_index(self, xy):

		"""
		desc:
			Converts a coordinate in pixels to a cell index. This allows you to
			determine on which widget a user has clicked.

		arguments:
			xy:
				desc:	An (x,y) coordinates tuple.
				type:	tuple

		returns:
			desc:	A cell index.
			type:	int
		"""

		for index in range(len(self.widgets)):
			x, y, w, h = self.get_rect(index)
			if x <= xy[0] and x+w >= xy[0] and y <= xy[1] and y+h >= xy[1]:
				return index
		return None

	def xy_to_widget(self, xy):

		index = self.xy_to_index(xy)
		if index is None:
			return None
		return self.widgets[index]


form = Form



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from libopensesame import plugins
from libopensesame.oslogging import oslogger
from libopensesame.item import item
from libqtopensesame.items.qtautoplugin import qtautoplugin

Quest = None
try:
	import Quest
	oslogger.debug(u'Loading Quest module directly')
except:
	oslogger.debug(u'Failed to load Quest module directly')
if Quest is None:
	try:
		from psychopy.contrib import quest as Quest
		oslogger.debug(u'Loading Quest module from PsychoPy')
	except:
		oslogger.debug(u'Failed to load Quest module from PsychoPy')
if Quest is None:
	try:
		Quest = plugins.load_mod(__file__, u'Quest')
		oslogger.debug(u'Loading Quest module from plug-in folder')
	except:
		oslogger.debug(u'Failed to load Quest module from plug-in folder')
if Quest is None:
		raise osexception(u'Failed to load Quest module.')

class quest_staircase_init(item):

	"""
	desc:
		A plug-in that iniializes a Quest staircase.
	"""

	description = u'Initializes a new Quest staircase procedure'

	def reset(self):

		"""
		desc:
			Initialize default variables.
		"""

		self.var.t_guess = .5
		self.var.t_guess_sd = .25
		self.var.p_threshold = .75
		self.var.beta = 3.5
		self.var.delta = .01
		self.var.gamma = .5
		self.var.test_value_method = u'quantile'
		self.var.min_test_value = 0
		self.var.max_test_value = 1
		self.var.var_test_value = u'quest_test_value'

	def quest_set_next_test_value(self):

		"""
		desc:
			Sets the next test value for the Quest procedure.
		"""

		if self.var.test_value_method == u'quantile':
			self.experiment.quest_test_value = self.experiment.quest.quantile
		elif self.var.test_value_method == u'mean':
			self.experiment.quest_test_value = self.experiment.quest.mean
		elif self.var.test_value_method == u'mode':
			self.experiment.quest_test_value = self.experiment.quest.mode
		else:
			raise osexception(
				u'Unknown test_value_method \'%s\' in quest_staircase_init' \
				% self.var.test_value_method)
		test_value = max(self.var.min_test_value, min(
			self.var.max_test_value, self.experiment.quest_test_value()))
		oslogger.debug(u'quest_test_value = %s' % test_value)
		self.experiment.var.quest_test_value = test_value
		self.experiment.var.set(self.var.var_test_value, test_value)

	def prepare(self):

		"""
		desc:
			Prepares the plug-in.
		"""

		self.experiment.quest = Quest.QuestObject(self.var.t_guess,
			self.var.t_guess_sd, self.var.p_threshold, self.var.beta,
			self.var.delta, self.var.gamma)
		self.experiment.quest_set_next_test_value = \
			self.quest_set_next_test_value
		self.experiment.quest_set_next_test_value()

	def var_info(self):

		"""
		desc:
			Gives a list of dictionaries with variable descriptions.

		returns:
			desc:	A list of (name, description) tuples.
			type:	list
		"""

		return item.var_info(self) + [(u'quest_test_value',
			u'(Determined by Quest procedure)')]

class qtquest_staircase_init(quest_staircase_init, qtautoplugin):

	"""
	desc:
		The GUI part of the plug-in. Controls are defined in info.json.
	"""

	def __init__(self, name, experiment, script=None):

		"""
		desc:
			Constructor.

		arguments:
			name:		The name of the plug-in.
			experiment:	The experiment object.

		keywords:
			script:		A definition script.
		"""

		quest_staircase_init.__init__(self, name, experiment, script)
		qtautoplugin.__init__(self, __file__)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.extensions import base_extension
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'quick_switcher', category=u'extension')


class quick_switcher(base_extension):

	"""
	desc:
		The quick-switcher allows you to quickly navigate to items.
	"""

	def activate(self):

		haystack = []
		for item in self.experiment.items.values():
			haystack.append((
				u'{} ({})'.format(item.name, item.item_type),
				item,
				item.open_tab
			))
		self.extension_manager.fire(
			u'quick_select',
			haystack=haystack,
			placeholder_text=_(u'Search items …')
		)
